# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent

from testtools.matchers import Contains, Equals
import fixtures

from snapcraft import storeapi

from . import FakeStoreCommandsBaseTestCase, get_sample_key


class ListKeysCommandTestCase(FakeStoreCommandsBaseTestCase):

    command_name = "list-keys"

    def test_command_without_login_must_ask(self):
        # TODO: look into why this many calls are done inside snapcraft.storeapi
        self.fake_store_account_info.mock.side_effect = [
            storeapi.http_clients.errors.InvalidCredentialsError("error"),
            {"account_id": "abcd", "account_keys": list()},
            {"account_id": "abcd", "account_keys": list()},
            {"account_id": "abcd", "account_keys": list()},
        ]

        result = self.run_command(
            [self.command_name], input="user@example.com\nsecret\n"
        )
        self.assertThat(
            result.output, Contains("You are required to login before continuing.")
        )

    def test_list_keys_successfully(self):
        self.fake_store_account_info.mock.return_value = {
            "account_id": "abcd",
            "account_keys": [
                {
                    "name": "default",
                    "public-key-sha3-384": (get_sample_key("default")["sha3-384"]),
                }
            ],
        }

        result = self.run_command(
            [self.command_name], input="user@example.com\nsecret\n"
        )

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Contains(
                dedent(
                    """\
                Name     SHA3-384 fingerprint
            *   default  {default_sha3_384}
            -   another  {another_sha3_384}  (not registered)
            """
                ).format(
                    default_sha3_384=get_sample_key("default")["sha3-384"],
                    another_sha3_384=get_sample_key("another")["sha3-384"],
                )
            ),
        )

    def test_list_keys_no_keys_on_system(self):
        self.fake_store_account_info.mock.return_value = {
            "account_id": "abcd",
            "account_keys": [
                {
                    "name": "default",
                    "public-key-sha3-384": (get_sample_key("default")["sha3-384"]),
                }
            ],
        }

        self.useFixture(
            fixtures.MockPatch("subprocess.check_output", return_value="[]".encode())
        )

        result = self.run_command(
            [self.command_name], input="user@example.com\nsecret\n"
        )

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Contains(
                dedent(
                    """\
                    No keys have been created on this system.  See 'snapcraft create-key --help' to create a key.
                    The following SHA3-384 key fingerprints have been registered but are not available on this system:
                    - vdEeQvRxmZ26npJCFaGnl-VfGz0lU2jZZkWp_s7E-RxVCNtH2_mtjcxq2NkDKkIp
            """
                )
            ),
        )

    def test_alias(self):
        self.command_name = "keys"
        self.fake_store_account_info.mock.return_value = {
            "account_id": "abcd",
            "account_keys": [
                {
                    "name": "default",
                    "public-key-sha3-384": (get_sample_key("default")["sha3-384"]),
                }
            ],
        }

        result = self.run_command(
            [self.command_name], input="user@example.com\nsecret\n"
        )

        self.assertThat(result.exit_code, Equals(0))

    def test_list_keys_without_registered(self):
        result = self.run_command(
            [self.command_name], input="user@example.com\nsecret\n"
        )

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output,
            Contains(
                "No keys have been registered. "
                "See 'snapcraft register-key --help' to register a key."
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import logging
import os
import stat
import textwrap
from unittest.mock import patch

import fixtures
import testtools
from testtools.matchers import (
    Annotate,
    Contains,
    DirExists,
    Equals,
    FileContains,
    FileExists,
    HasLength,
    Not,
)

from snapcraft import extractors, yaml_utils
from snapcraft.internal import errors, project_loader, states
from snapcraft.internal.meta import _snap_packaging
from snapcraft.internal.meta import errors as meta_errors
from snapcraft.project import Project
from tests import fixture_setup, unit


class CreateBaseTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.config_data = {
            "architectures": [{"build-on": "all", "run-on": "amd64"}],
            "name": "my-package",
            "base": "core18",
            "version": "1.0",
            "description": "my description",
            "summary": "my summary",
            "confinement": "devmode",
            "environment": {"GLOBAL": "y"},
            "parts": {"test-part": {"plugin": "nil"}},
        }

        self.snapcraft_yaml_file_path = os.path.join("snap", "snapcraft.yaml")
        # Ensure the ensure snapcraft.yaml method has something to copy.
        _create_file(self.snapcraft_yaml_file_path)

    def generate_meta_yaml(
        self, *, build=False, actual_prime_dir=None, snapcraft_yaml_file_path=None
    ):
        if snapcraft_yaml_file_path is None:
            snapcraft_yaml_file_path = self.snapcraft_yaml_file_path
        os.makedirs("snap", exist_ok=True)
        with open(snapcraft_yaml_file_path, "w") as f:
            f.write(yaml_utils.dump(self.config_data))

        self.project = Project(snapcraft_yaml_file_path=snapcraft_yaml_file_path)
        if actual_prime_dir is not None:
            self.project._prime_dir = actual_prime_dir

        self.meta_dir = os.path.join(self.project.prime_dir, "meta")
        self.hooks_dir = os.path.join(self.meta_dir, "hooks")
        self.snap_yaml = os.path.join(self.meta_dir, "snap.yaml")

        self.config = project_loader.load_config(project=self.project)
        if build:
            for part in self.config.parts.all_parts:
                part.pull()
                part.build()
                part.stage()
                part.prime()

        _snap_packaging.create_snap_packaging(self.config)

        self.assertTrue(os.path.exists(self.snap_yaml), "snap.yaml was not created")

        with open(self.snap_yaml) as f:
            return yaml_utils.load(f)


class CreateTestCase(CreateBaseTestCase):
    def test_create_meta(self):
        y = self.generate_meta_yaml()

        expected = {
            "architectures": ["amd64"],
            "confinement": "devmode",
            "grade": "stable",
            "description": "my description",
            "environment": {"GLOBAL": "y"},
            "summary": "my summary",
            "name": "my-package",
            "base": "core18",
            "version": "1.0",
        }

        self.assertThat(y, Equals(expected))

    def test_create_meta_with_core_as_base(self):
        self.config_data["base"] = "core"
        self.fake_snapd.snaps_result = [
            dict(name="core", channel="stable", revision="10")
        ]

        y = self.generate_meta_yaml()

        self.assertThat(y, Not(Contains("base")))

    def test_create_meta_default_architecture(self):
        del self.config_data["architectures"]

        y = self.generate_meta_yaml()

        expected = {
            "architectures": [self.project.deb_arch],
            "confinement": "devmode",
            "grade": "stable",
            "description": "my description",
            "environment": {"GLOBAL": "y"},
            "summary": "my summary",
            "name": "my-package",
            "base": "core18",
            "version": "1.0",
        }

        self.assertThat(y, Equals(expected))

    def test_create_meta_with_epoch(self):
        self.config_data["epoch"] = "1*"

        y = self.generate_meta_yaml()
        self.assertTrue(
            "epoch" in y, 'Expected "epoch" property to be copied into snap.yaml'
        )
        self.assertThat(y["epoch"], Equals("1*"))

    def test_create_meta_with_title(self):
        self.config_data["title"] = "The Title"

        y = self.generate_meta_yaml()
        self.assertThat(y, Contains("title"))
        self.assertThat(y["title"], Equals("The Title"))

    def test_create_meta_with_license(self):
        self.config_data["license"] = "MIT"

        y = self.generate_meta_yaml()
        self.assertThat(y, Contains("license"))
        self.assertThat(y["license"], Equals("MIT"))

    def test_create_meta_with_assumes(self):
        self.config_data["assumes"] = ["feature1", "feature2"]

        y = self.generate_meta_yaml()
        self.assertTrue(
            "assumes" in y, 'Expected "assumes" property to be copied into snap.yaml'
        )
        self.assertThat(y["assumes"], Equals(["feature1", "feature2"]))

    def test_create_meta_command_chain_with_assumes(self):
        self.config_data["assumes"] = ["feature1", "feature2"]
        self.config_data["apps"] = {"app": {"command": "foo", "command-chain": ["bar"]}}
        _create_file(os.path.join(self.prime_dir, "foo"), executable=True)
        _create_file(os.path.join(self.prime_dir, "bar"), executable=True)

        y = self.generate_meta_yaml()
        self.assertTrue(
            "assumes" in y, 'Expected "assumes" property to be copied into snap.yaml'
        )
        self.assertThat(
            y["assumes"], Equals(sorted(["feature1", "feature2", "command-chain"]))
        )

    def test_create_gadget_meta_with_gadget_yaml(self):
        gadget_yaml = "stub entry: stub value"
        _create_file("gadget.yaml", content=gadget_yaml)

        self.config_data["type"] = "gadget"

        self.generate_meta_yaml()

        expected_gadget = os.path.join(self.meta_dir, "gadget.yaml")
        self.assertTrue(os.path.exists(expected_gadget))

        self.assertThat(expected_gadget, FileContains(gadget_yaml))

    def test_create_gadget_meta_with_missing_gadget_yaml_raises_error(self):
        self.config_data["type"] = "gadget"

        self.assertRaises(errors.MissingGadgetError, self.generate_meta_yaml)

    def test_create_kernel_meta_with_kernel_yaml(self):
        kernel_yaml = "stub entry: stub value"
        _create_file("kernel.yaml", content=kernel_yaml)

        self.config_data["type"] = "kernel"
        self.config_data["build-base"] = "core20"
        self.config_data.pop("base")

        self.generate_meta_yaml()

        expected_kernel = os.path.join(self.meta_dir, "kernel.yaml")
        self.assertTrue(os.path.exists(expected_kernel))

        self.assertThat(expected_kernel, FileContains(kernel_yaml))

    def test_create_meta_with_declared_icon(self):
        _create_file(os.path.join(os.curdir, "my-icon.png"))
        self.config_data["icon"] = "my-icon.png"

        y = self.generate_meta_yaml()

        self.assertThat(os.path.join(self.meta_dir, "gui", "icon.png"), FileExists())

        self.assertFalse("icon" in y, "icon found in snap.yaml {}".format(y))

    def test_create_meta_with_declared_icon_with_dots(self):
        _create_file("com.my.icon.png")
        self.config_data["icon"] = "com.my.icon.png"

        y = self.generate_meta_yaml()

        self.assertThat(os.path.join(self.meta_dir, "gui", "icon.png"), FileExists())

        self.assertFalse("icon" in y, "icon found in snap.yaml {}".format(y))

    def test_create_meta_with_declared_icon_in_parent_dir(self):
        _create_file("my-icon.png")
        builddir = os.path.join(os.curdir, "subdir")
        os.mkdir(builddir)
        os.chdir(builddir)
        self.config_data["icon"] = "../my-icon.png"

        y = self.generate_meta_yaml()

        self.assertThat(os.path.join(self.meta_dir, "gui", "icon.png"), FileExists())

        self.assertFalse("icon" in y, "icon found in snap.yaml {}".format(y))

    def test_create_meta_with_declared_icon_and_setup(self):
        fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(fake_logger)

        gui_path = os.path.join("setup", "gui")
        os.makedirs(gui_path)
        setup_icon_content = "setup icon"
        _create_file(os.path.join(gui_path, "icon.png"), content=setup_icon_content)

        declared_icon_content = "declared icon"
        _create_file("my-icon.png", content=declared_icon_content)
        self.config_data["icon"] = "my-icon.png"

        y = self.generate_meta_yaml()

        expected_icon = os.path.join(self.meta_dir, "gui", "icon.png")
        self.assertTrue(
            os.path.exists(expected_icon), "icon.png was not setup correctly"
        )
        self.assertThat(expected_icon, FileContains(declared_icon_content))

        self.assertFalse("icon" in y, "icon found in snap.yaml {}".format(y))

        # Check for the correct deprecation message.
        self.assertIn(
            "Assets in 'setup/gui' should now be placed in 'snap/gui'.",
            fake_logger.output,
        )
        self.assertIn(
            "See http://snapcraft.io/docs/deprecation-notices/dn3", fake_logger.output
        )

    @patch("os.link", side_effect=OSError("Invalid cross-device link"))
    def test_create_meta_with_declared_icon_when_os_link_raises(self, link_mock):
        _create_file(os.path.join(os.curdir, "my-icon.png"))
        self.config_data["icon"] = "my-icon.png"

        y = self.generate_meta_yaml()

        self.assertThat(os.path.join(self.meta_dir, "gui", "icon.png"), FileExists())

        self.assertFalse("icon" in y, "icon found in snap.yaml {}".format(y))

    def test_create_meta_with_declared_icon_and_setup_ran_twice_ok(self):
        gui_path = os.path.join("setup", "gui")
        os.makedirs(gui_path)
        icon_content = "setup icon"
        _create_file(os.path.join(gui_path, "icon.png"), content=icon_content)

        _create_file("my-icon.png")
        self.config_data["icon"] = "my-icon.png"

        self.generate_meta_yaml()

        # Running again should be good
        _snap_packaging.create_snap_packaging(self.config)

    def test_create_meta_with_icon_in_setup(self):
        gui_path = os.path.join("setup", "gui")
        os.makedirs(gui_path)
        icon_content = "setup icon"
        _create_file(os.path.join(gui_path, "icon.png"), content=icon_content)

        y = self.generate_meta_yaml()

        expected_icon = os.path.join(self.meta_dir, "gui", "icon.png")
        self.assertThat(expected_icon, FileContains(icon_content))

        self.assertFalse("icon" in y, "icon found in snap.yaml {}".format(y))

    def test_version_script(self):
        self.config_data["version-script"] = "echo 10.1-devel"

        y = self.generate_meta_yaml()

        self.assertThat(y["version"], Equals("10.1-devel"))

    def test_version_script_exits_bad(self):
        self.config_data["version-script"] = "exit 1"

        with testtools.ExpectedException(meta_errors.CommandError):
            self.generate_meta_yaml()

    def test_version_script_with_no_output(self):
        self.config_data["version-script"] = "echo"

        with testtools.ExpectedException(meta_errors.CommandError):
            self.generate_meta_yaml()

    def test_layout(self):
        layout = {"/target": {"bind": "$SNAP/foo"}}
        self.config_data["layout"] = layout

        y = self.generate_meta_yaml()

        self.assertThat(y["layout"], Equals(layout))

    def test_adapter_full_with_command_chain(self):
        self.config_data["apps"] = {"app": {"command": "foo", "command-chain": ["bar"]}}
        _create_file(os.path.join(self.prime_dir, "foo"), executable=True)
        _create_file(os.path.join(self.prime_dir, "bar"), executable=True)

        y = self.generate_meta_yaml()

        self.expectThat(y["apps"]["app"]["command"], Equals("foo"))
        self.expectThat(
            y["apps"]["app"]["command-chain"],
            Equals([os.path.join("snap", "command-chain", "snapcraft-runner"), "bar"]),
        )
        self.assertThat(y["assumes"], Equals(["command-chain"]))

    def test_adapter_assumes_command_chain(self):
        self.config_data["apps"] = {"app": {"command": "foo"}}
        _create_file(os.path.join(self.prime_dir, "foo"), executable=True)

        y = self.generate_meta_yaml()

        self.expectThat(y["apps"]["app"]["command"], Equals("foo"))
        self.expectThat(
            y["apps"]["app"]["command-chain"],
            Equals([os.path.join("snap", "command-chain", "snapcraft-runner")]),
        )
        self.assertThat(y["assumes"], Equals(["command-chain"]))

    def test_adapter_none(self):
        # Adapter "none" will passthrough command as-is.
        self.config_data["apps"] = {"app": {"adapter": "none", "command": "/foo"}}

        y = self.generate_meta_yaml()

        self.assertThat(y["apps"]["app"]["command"], Equals("/foo"))
        self.assertThat(y["apps"]["app"].get("command-chain"), Equals(None))


class StopModeTestCase(CreateBaseTestCase):

    stop_modes = [
        "sigterm",
        "sigterm-all",
        "sighup",
        "sighup-all",
        "sigusr1",
        "sigusr1-all",
        "sigusr2",
        "sigusr2-all",
    ]

    def test_valid(self):
        for mode in self.stop_modes:
            self.config_data["apps"] = {
                "app1": {"command": "sh", "daemon": "simple", "stop-mode": mode}
            }

            y = self.generate_meta_yaml()

            self.expectThat(y["apps"]["app1"]["stop-mode"], Equals(mode))


class RefreshModeTestCase(CreateBaseTestCase):

    refresh_modes = ["endure", "restart"]

    def test_valid(self):
        for refresh_mode in self.refresh_modes:
            self.config_data["apps"] = {
                "app1": {
                    "command": "sh",
                    "daemon": "simple",
                    "refresh-mode": refresh_mode,
                }
            }

            y = self.generate_meta_yaml()

            self.expectThat(y["apps"]["app1"]["refresh-mode"], Equals(refresh_mode))


class BeforeAndAfterTest(CreateBaseTestCase):
    def test_before_valid(self):
        self.config_data["apps"] = {
            "app-before": {"command": "sh", "daemon": "simple"},
            "app": {"command": "sh", "daemon": "simple", "before": ["app-before"]},
        }

        y = self.generate_meta_yaml()

        self.assertThat(y["apps"]["app"]["before"], Equals(["app-before"]))

    def test_after_valid(self):
        self.config_data["apps"] = {
            "app-after": {"command": "sh", "daemon": "simple"},
            "app": {"command": "sh", "daemon": "simple", "after": ["app-after"]},
        }

        y = self.generate_meta_yaml()

        self.assertThat(y["apps"]["app"]["after"], Equals(["app-after"]))


class PassthroughBaseTestCase(CreateBaseTestCase):
    def setUp(self):
        super().setUp()

        self.config_data = {
            "name": "my-package",
            "base": "core18",
            "version": "1.0",
            "grade": "stable",
            "description": "my description",
            "summary": "my summary",
            "parts": {"test-part": {"plugin": "nil"}},
        }


class PassthroughErrorTestCase(PassthroughBaseTestCase):
    def test_ambiguous_key_fails(self):
        self.config_data["confinement"] = "devmode"
        self.config_data["passthrough"] = {"confinement": "next-generation"}
        raised = self.assertRaises(
            meta_errors.AmbiguousPassthroughKeyError, self.generate_meta_yaml
        )
        self.assertThat(raised.keys, Equals("'confinement'"))

    def test_app_ambiguous_key_fails(self):
        self.config_data["apps"] = {
            "foo": {
                "command": "echo",
                "daemon": "simple",
                "passthrough": {"daemon": "complex"},
            }
        }
        raised = self.assertRaises(
            meta_errors.AmbiguousPassthroughKeyError, self.generate_meta_yaml
        )
        self.assertThat(raised.keys, Equals("'daemon'"))

    def test_hook_ambiguous_key_fails(self):
        self.config_data["hooks"] = {
            "foo": {"plugs": ["network"], "passthrough": {"plugs": ["network"]}}
        }
        raised = self.assertRaises(
            meta_errors.AmbiguousPassthroughKeyError, self.generate_meta_yaml
        )
        self.assertThat(raised.keys, Equals("'plugs'"))

    def test_warn_once_only(self):
        fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(fake_logger)

        self.config_data["confinement"] = "devmode"
        self.config_data["passthrough"] = {"foo": "bar", "spam": "eggs"}
        self.config_data["apps"] = {
            "foo": {"command": "echo", "passthrough": {"foo": "bar", "spam": "eggs"}}
        }
        self.config_data["hooks"] = {
            "foo": {"plugs": ["network"], "passthrough": {"foo": "bar", "spam": "eggs"}}
        }
        self.generate_meta_yaml()

        output_lines = fake_logger.output.splitlines()
        self.assertThat(
            output_lines,
            Contains(
                "The 'passthrough' property is being used to propagate "
                "experimental properties to snap.yaml that have not been "
                "validated."
            ),
        )
        self.assertThat(
            len(output_lines),
            Annotate(
                "There were duplicate lines logged.", Equals(len(set(output_lines)))
            ),
        )


class PassthroughPropagateTestCase(PassthroughBaseTestCase):

    cases = [
        (
            "new",
            dict(
                snippet={"passthrough": {"spam": "eggs"}},
                section=None,
                name=None,
                key="spam",
                value="eggs",
            ),
        ),
        (
            "different type",
            dict(
                # This is normally an array of strings
                snippet={"passthrough": {"architectures": "all"}},
                section=None,
                name=None,
                key="architectures",
                value="all",
            ),
        ),
        (
            "with default",
            dict(
                snippet={"passthrough": {"confinement": "next-generation"}},
                section=None,
                name=None,
                key="confinement",
                value="next-generation",
            ),
        ),
        (
            "app, new",
            dict(
                snippet={
                    "apps": {
                        "foo": {"command": "echo", "passthrough": {"spam": "eggs"}}
                    }
                },
                section="apps",
                name="foo",
                key="spam",
                value="eggs",
            ),
        ),
        (
            "app, different type",
            dict(
                snippet={
                    "apps": {
                        "foo": {
                            "command": "echo",
                            # This is normally an array of strings
                            "passthrough": {"aliases": "foo"},
                        }
                    }
                },
                section="apps",
                name="foo",
                key="aliases",
                value="foo",
            ),
        ),
        # Note: There are currently no app properties with defaults
        (
            "hook, new",
            dict(
                snippet={
                    "hooks": {
                        "foo": {"plugs": ["network"], "passthrough": {"spam": "eggs"}}
                    }
                },
                section="hooks",
                name="foo",
                key="spam",
                value="eggs",
            ),
        ),
        (
            "hook, different type",
            dict(
                snippet={
                    "hooks": {
                        "foo": {
                            # This is normally an array of strings
                            "passthrough": {"plugs": "network"}
                        }
                    }
                },
                section="hooks",
                name="foo",
                key="plugs",
                value="network",
            ),
        ),
        # Note: There are currently no hook properties with defaults
    ]

    def assert_passthrough(self, snippet, section, name, key, value):
        fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(fake_logger)

        self.config_data.update(snippet)
        y = self.generate_meta_yaml()
        if section:
            y = y[section][name]
            self.expectThat(
                y,
                Contains(key),
                "Expected {!r} property to be propagated to snap.yaml".format(key),
            )
            self.expectThat(y[key], Equals(value))
            self.expectThat(
                fake_logger.output,
                Contains(
                    "The 'passthrough' property is being used to propagate "
                    "experimental properties to snap.yaml that have not been "
                    "validated.\n"
                ),
            )

    def test_propagate(self):
        for case in self.cases:
            self.assert_passthrough(**case[1])


class CreateMetadataFromSourceBaseTestCase(CreateBaseTestCase):
    def setUp(self):
        super().setUp()
        self.config_data = {
            "name": "test-name",
            "base": "core18",
            "version": "test-version",
            "summary": "test-summary",
            "description": "test-description",
            "adopt-info": "test-part",
            "parts": {
                "test-part": {
                    "plugin": "dump",
                    "source": ".",
                    "parse-info": ["test-metadata-file"],
                }
            },
            "apps": {"test-app": {"command": "echo"}},
        }
        # Create metadata file
        open("test-metadata-file", "w").close()


class CreateMetadataFromSourceTestCase(CreateMetadataFromSourceBaseTestCase):
    def test_create_metadata_with_missing_parse_info(self):
        del self.config_data["summary"]
        del self.config_data["parts"]["test-part"]["parse-info"]
        raised = self.assertRaises(
            meta_errors.AdoptedPartNotParsingInfo, self.generate_meta_yaml, build=True
        )
        self.assertThat(raised.part, Equals("test-part"))

    def test_create_metadata_with_wrong_adopt_info(self):
        del self.config_data["summary"]
        self.config_data["adopt-info"] = "wrong-part"
        raised = self.assertRaises(
            meta_errors.AdoptedPartMissingError, self.generate_meta_yaml
        )
        self.assertThat(raised.part, Equals("wrong-part"))

    def test_metadata_doesnt_overwrite_specified(self):
        fake_logger = fixtures.FakeLogger(level=logging.WARNING)
        self.useFixture(fake_logger)

        def _fake_extractor(file_path, workdir):
            return extractors.ExtractedMetadata(
                summary="extracted summary", description="extracted description"
            )

        self.useFixture(fixture_setup.FakeMetadataExtractor("fake", _fake_extractor))

        y = self.generate_meta_yaml(build=True)

        # Since both summary and description were specified, neither should be
        # overwritten
        self.assertThat(y["summary"], Equals(self.config_data["summary"]))
        self.assertThat(y["description"], Equals(self.config_data["description"]))

        # Verify that we warn that the YAML took precedence over the extracted
        # metadata for summary and description
        self.assertThat(
            fake_logger.output,
            Contains(
                "The 'description' and 'summary' properties are specified in "
                "adopted info as well as the YAML: taking the properties from the "
                "YAML"
            ),
        )

    def test_metadata_with_unexisting_icon(self):
        def _fake_extractor(file_path, workdir):
            return extractors.ExtractedMetadata(
                icon="test/extracted/unexistent/icon/path"
            )

        self.useFixture(fixture_setup.FakeMetadataExtractor("fake", _fake_extractor))

        # The meta generation should just ignore the dead path, and not fail.
        self.generate_meta_yaml(build=True)

    def test_metadata_satisfies_required_property(self):
        del self.config_data["summary"]

        def _fake_extractor(file_path, workdir):
            return extractors.ExtractedMetadata(
                summary="extracted summary", description="extracted description"
            )

        self.useFixture(fixture_setup.FakeMetadataExtractor("fake", _fake_extractor))

        y = self.generate_meta_yaml(build=True)

        # Summary should come from the extracted metadata, while description
        # should not.
        self.assertThat(y["summary"], Equals("extracted summary"))
        self.assertThat(y["description"], Equals(self.config_data["description"]))

    def test_metadata_not_all_properties_satisfied(self):
        del self.config_data["summary"]
        del self.config_data["description"]

        def _fake_extractor(file_path, workdir):
            return extractors.ExtractedMetadata(description="extracted description")

        self.useFixture(fixture_setup.FakeMetadataExtractor("fake", _fake_extractor))

        # Assert that description has been satisfied by extracted metadata, but
        # summary has not.
        raised = self.assertRaises(
            meta_errors.MissingSnapcraftYamlKeysError,
            self.generate_meta_yaml,
            build=True,
        )
        self.assertThat(raised.keys, Equals("'summary'"))


class CreateWithAssetsTestCase(CreateBaseTestCase):
    def assert_create_meta_with_hook(self, snapcraft_assets_dir):
        hooksdir = os.path.join(snapcraft_assets_dir, "hooks")
        os.makedirs(hooksdir)
        _create_file(os.path.join(hooksdir, "foo"), executable=True)
        _create_file(os.path.join(hooksdir, "bar"), executable=True)
        self.config_data["hooks"] = {"foo": {"plugs": ["plug"]}, "bar": {}}

        y = self.generate_meta_yaml(
            snapcraft_yaml_file_path=os.path.join(
                snapcraft_assets_dir, "snapcraft.yaml"
            )
        )

        self.assertThat(
            y, Contains("hooks"), "Expected generated YAML to contain 'hooks'"
        )

        for hook in ("foo", "bar"):
            meta_dir_hook_path = os.path.join(self.prime_dir, "meta", "hooks", hook)
            snap_dir_hook_path = os.path.join(self.prime_dir, "snap", "hooks", hook)

            for hook_location in (meta_dir_hook_path, snap_dir_hook_path):
                self.assertThat(
                    hook_location,
                    FileExists(),
                    "The {!r} hook was not setup correctly".format(hook),
                )

        for hook in ("foo", "bar"):
            self.assertThat(
                y["hooks"],
                Contains(hook),
                "Expected generated hooks to contain {!r}".format(hook),
            )

        self.assertThat(
            y["hooks"]["foo"],
            Contains("plugs"),
            "Expected generated 'foo' hook to contain 'plugs'",
        )
        self.assertThat(y["hooks"]["foo"]["plugs"], HasLength(1))
        self.assertThat(y["hooks"]["foo"]["plugs"][0], Equals("plug"))
        self.assertThat(
            y["hooks"]["bar"],
            Not(Contains("plugs")),
            "Expected generated 'bar' hook to not contain 'plugs'",
        )

    def test_create_meta_with_hook_in_snap(self):
        self.assert_create_meta_with_hook("snap")

    def test_create_meta_with_hook_in_build_aux(self):
        self.assert_create_meta_with_hook("build-aux/snap")

    def assert_local_is_not_copied_to_snap(self, snapcraft_assets_dir):
        project_local_dir = os.path.join(snapcraft_assets_dir, "local")
        local_file = "file"
        local_subdir_file = os.path.join("dir", "file")

        os.makedirs(os.path.join(project_local_dir, "dir"))
        _create_file(os.path.join(project_local_dir, local_file))
        _create_file(os.path.join(project_local_dir, local_subdir_file))

        self.generate_meta_yaml(
            snapcraft_yaml_file_path=os.path.join(
                snapcraft_assets_dir, "snapcraft.yaml"
            )
        )

        prime_local_dir = os.path.join(self.prime_dir, "snap", "local")
        self.assertThat(os.path.join(prime_local_dir, local_file), Not(FileExists()))
        self.assertThat(
            os.path.join(prime_local_dir, local_subdir_file), Not(FileExists())
        )

    def test_local_is_not_copied_to_snap_with_snap(self):
        self.assert_create_meta_with_hook("snap")

    def test_local_is_not_copied_to_snap_with_build_aux(self):
        self.assert_create_meta_with_hook("build-aux/snap")


class MetadataFromSourceWithIconFileTestCase(CreateMetadataFromSourceBaseTestCase):
    def assert_no_overwrite(self, snapcraft_assets_dir, directory, file_name):
        os.makedirs(directory)
        icon_content = "setup icon"
        _create_file(os.path.join(directory, file_name), content=icon_content)

        def _fake_extractor(file_path, workdir):
            return extractors.ExtractedMetadata(
                icon="test/extracted/unexistent/icon/path"
            )

        self.useFixture(fixture_setup.FakeMetadataExtractor("fake", _fake_extractor))

        self.generate_meta_yaml(
            build=True,
            snapcraft_yaml_file_path=os.path.join(
                snapcraft_assets_dir, "snapcraft.yaml"
            ),
        )

        expected_icon = os.path.join(self.meta_dir, "gui", file_name)
        self.assertThat(expected_icon, FileContains(icon_content))

    def test_setup_gui_svg(self):
        self.assert_no_overwrite("snap", "setup/gui", "icon.svg")

    def test_snap_gui_svg(self):
        self.assert_no_overwrite("snap", "snap/gui", "icon.svg")

    def test_build_aux_gui_svg(self):
        self.assert_no_overwrite("build-aux/snap", "build-aux/snap/gui", "icon.svg")

    def test_setup_gui_png(self):
        self.assert_no_overwrite("snap", "setup/gui", "icon.png")

    def test_snap_gui_png(self):
        self.assert_no_overwrite("snap", "snap/gui", "icon.png")

    def test_build_aux_gui_png(self):
        self.assert_no_overwrite("build-aux/snap", "build-aux/snap/gui", "icon.png")


class MetadataFromSourceWithDesktopFileTestCase(CreateMetadataFromSourceBaseTestCase):
    def assert_no_overwrite(self, snapcraft_assets_dir, directory):
        os.makedirs(directory)
        desktop_content = "setup desktop"
        _create_file(
            os.path.join(directory, "test-app.desktop"), content=desktop_content
        )

        def _fake_extractor(file_path, workdir):
            return extractors.ExtractedMetadata(
                desktop_file_paths=[
                    "usr/share/applications/com.example.test/app.desktop"
                ]
            )

        self.useFixture(fixture_setup.FakeMetadataExtractor("fake", _fake_extractor))

        self.generate_meta_yaml(
            build=True,
            snapcraft_yaml_file_path=os.path.join(
                snapcraft_assets_dir, "snapcraft.yaml"
            ),
        )

        expected_desktop = os.path.join(self.meta_dir, "gui", "test-app.desktop")
        self.assertThat(expected_desktop, FileContains(desktop_content))

    def test_setup_gui(self):
        self.assert_no_overwrite("snap", "setup/gui")

    def test_snap_gui(self):
        self.assert_no_overwrite("snap", "snap/gui")

    def test_build_aux_gui(self):
        self.assert_no_overwrite("build-aux/snap", "build-aux/snap/gui")


class ScriptletsMetadataTestCase(CreateMetadataFromSourceBaseTestCase):
    def assert_scriptlets_satisfy_required_property(
        self, keyword, original, value, setter
    ):
        with contextlib.suppress(KeyError):
            del self.config_data[keyword]

        del self.config_data["parts"]["test-part"]["parse-info"]
        self.config_data["parts"]["test-part"][
            "override-prime"
        ] = "snapcraftctl {} {}".format(setter, value)

        generated = self.generate_meta_yaml(build=True)

        self.assertThat(generated[keyword], Equals(value))

    def assert_scriptlets_no_overwrite_existing_property(
        self, keyword, original, value, setter
    ):
        self.config_data[keyword] = original
        fake_logger = fixtures.FakeLogger(level=logging.WARNING)
        self.useFixture(fake_logger)

        del self.config_data["parts"]["test-part"]["parse-info"]
        self.config_data["parts"]["test-part"][
            "override-prime"
        ] = "snapcraftctl {} {}".format(setter, value)

        generated = self.generate_meta_yaml(build=True)

        self.assertThat(generated[keyword], Equals(original))

        # Since the specified version took precedence over the scriptlet-set
        # version, verify that we warned
        self.assertThat(
            fake_logger.output,
            Contains(
                "The {!r} property is specified in adopted info as well as "
                "the YAML: taking the property from the YAML".format(keyword)
            ),
        )

    def assert_scriptlets_overwrite_extracted_metadata(
        self, keyword, original, value, setter
    ):
        with contextlib.suppress(KeyError):
            del self.config_data[keyword]

        self.config_data["parts"]["test-part"][
            "override-build"
        ] = "snapcraftctl build && snapcraftctl {} {}".format(setter, value)

        def _fake_extractor(file_path, workdir):
            return extractors.ExtractedMetadata(**{keyword: "extracted-value"})

        self.useFixture(fixture_setup.FakeMetadataExtractor("fake", _fake_extractor))

        generated = self.generate_meta_yaml(build=True)

        self.assertThat(generated[keyword], Equals(value))

    def assert_scriptlets_overwrite_extracted_metadata_regardless_of_order(
        self, keyword, original, value, setter
    ):
        with contextlib.suppress(KeyError):
            del self.config_data[keyword]

        self.config_data["parts"]["test-part"][
            "override-pull"
        ] = "snapcraftctl {} {} && snapcraftctl pull".format(setter, value)

        def _fake_extractor(file_path, workdir):
            return extractors.ExtractedMetadata(**{keyword: "extracted-value"})

        self.useFixture(fixture_setup.FakeMetadataExtractor("fake", _fake_extractor))

        generated = self.generate_meta_yaml(build=True)

        self.assertThat(generated[keyword], Equals(value))

    def test_scriptlets_satisfy_required_property_for_grade(self):
        self.assert_scriptlets_satisfy_required_property(
            "grade", "stable", "devel", "set-grade"
        )

    def test_scriptlets_overwrite_extracted_metadata_for_grade(self):
        self.assert_scriptlets_overwrite_extracted_metadata(
            "grade", "stable", "devel", "set-grade"
        )

    def test_scriptlets_no_overwrite_existing_property_for_grade(self):
        self.assert_scriptlets_no_overwrite_existing_property(
            "grade", "stable", "devel", "set-grade"
        )

    def test_scriptlets_overwrite_extracted_metadata_regardless_of_order_for_grade(
        self,
    ):
        self.assert_scriptlets_overwrite_extracted_metadata_regardless_of_order(
            "grade", "stable", "devel", "set-grade"
        )

    def test_scriptlets_satisfy_required_property_for_version(self):
        self.assert_scriptlets_satisfy_required_property(
            "version", "original-version", "new-version", "set-version"
        )

    def test_scriptlets_overwrite_extracted_metadata_for_version(self):
        self.assert_scriptlets_overwrite_extracted_metadata(
            "version", "original-version", "new-version", "set-version"
        )

    def test_scriptlets_no_overwrite_existing_property_for_version(self):
        self.assert_scriptlets_no_overwrite_existing_property(
            "version", "original-version", "new-version", "set-version"
        )

    def test_scriptlets_overwrite_extracted_metadata_regardless_of_order_for_version(
        self,
    ):
        self.assert_scriptlets_overwrite_extracted_metadata_regardless_of_order(
            "version", "original-version", "new-version", "set-version"
        )


class WriteSnapDirectoryTestCase(CreateBaseTestCase):
    def test_write_snap_directory(self):
        # Setup a snap directory containing a few things.
        _create_file(os.path.join(self.snap_dir, "snapcraft.yaml"))
        _create_file(os.path.join(self.snap_dir, "hooks", "test-hook"), executable=True)

        # Now write the snap directory, and verify everything was migrated, as
        # well as the hook making it into meta/.
        self.generate_meta_yaml()
        prime_snap_dir = os.path.join(self.prime_dir, "snap")

        self.assertThat(
            os.path.join(prime_snap_dir, "hooks", "test-hook"), FileExists()
        )
        self.assertThat(os.path.join(self.hooks_dir, "test-hook"), FileExists())

        # The hook should be empty, because the one in snap/hooks is empty, and
        # no wrapper is generated (i.e. that hook is copied to both locations).
        self.assertThat(os.path.join(self.hooks_dir, "test-hook"), FileContains(""))

    def test_nothing_to_write(self):
        # Setup a snap directory containing just a snapcraft.yaml, nothing else
        _create_file(os.path.join(self.snap_dir, "snapcraft.yaml"))

        # Now generate the metadata, and verify that no snap directory was
        # created
        self.generate_meta_yaml()
        self.assertThat(
            os.path.join(self.prime_dir, "snap"),
            Not(DirExists()),
            "Expected snap directory to NOT be created",
        )

    def test_snap_hooks_overwrite_part_hooks(self):
        # Setup a prime/snap directory containing a hook.
        part_hook = os.path.join(self.prime_dir, "snap", "hooks", "test-hook")
        _create_file(part_hook, content="from part", executable=True)

        # Setup a snap directory containing the same hook
        snap_hook = os.path.join(self.snap_dir, "hooks", "test-hook")
        _create_file(snap_hook, content="from snap", executable=True)

        # Now write the snap directory, and verify that the snap hook overwrote
        # the part hook in both prime/snap/hooks and prime/meta/hooks.
        self.generate_meta_yaml()
        prime_snap_dir = os.path.join(self.prime_dir, "snap")
        self.assertThat(
            os.path.join(prime_snap_dir, "hooks", "test-hook"), FileExists()
        )
        self.assertThat(os.path.join(self.hooks_dir, "test-hook"), FileExists())

        # Both hooks in snap/hooks and meta/hooks should contain 'from snap' as
        # that one should have overwritten the other (and its wrapper).
        self.assertThat(
            os.path.join(self.prime_dir, "snap", "hooks", "test-hook"),
            FileContains("from snap"),
        )
        self.assertThat(
            os.path.join(self.prime_dir, "meta", "hooks", "test-hook"),
            FileContains("from snap"),
        )

    def test_hooks_with_command_chain_write_stub(self):
        self.config_data["hooks"] = {
            "configure": {"command-chain": ["fake-hook-command-chain"]}
        }

        self.generate_meta_yaml()

        self.assertThat(os.path.join(self.hooks_dir, "configure"), FileExists())

        # The hook should be empty, because the one in snap/hooks is empty, and
        # no wrapper is generated (i.e. that hook is copied to both locations).
        self.assertThat(
            os.path.join(self.hooks_dir, "configure"), FileContains("#!/bin/sh\n")
        )

    def test_snap_hooks_stubs_not_created_stubs_from_command_chain(self):
        self.config_data["hooks"] = {
            "install": {"command-chain": ["fake-hook-command-chain"]}
        }

        # Setup a prime/snap directory containing a hook.
        part_hook = os.path.join(self.prime_dir, "snap", "hooks", "install")
        _create_file(part_hook, content="from part", executable=True)

        # Now write the snap directory, and verify that the snap hook overwrote
        # the part hook in both prime/snap/hooks and prime/meta/hooks.
        self.generate_meta_yaml()

        self.assertThat(
            os.path.join(self.hooks_dir, "install"),
            FileContains(
                textwrap.dedent(
                    """\
            #!/bin/sh
            export PATH="$SNAP/usr/sbin:$SNAP/usr/bin:$SNAP/sbin:$SNAP/bin${PATH:+:$PATH}"
            export LD_LIBRARY_PATH="$SNAP_LIBRARY_PATH${LD_LIBRARY_PATH:+:$LD_LIBRARY_PATH}"
            exec "$SNAP/snap/hooks/install" "$@"
            """
                )
            ),
        )

    def test_snap_hooks_not_executable_chmods(self):
        real_chmod = os.chmod

        def chmod_noop(name, *args, **kwargs):
            # Simulate a source filesystem that doesn't allow changing permissions
            if name.startswith("snap"):
                pass
            else:
                real_chmod(name, *args, **kwargs)

        self.useFixture(fixtures.MonkeyPatch("os.chmod", chmod_noop))

        # Setup a snap directory containing a few things.
        _create_file(os.path.join(self.snap_dir, "snapcraft.yaml"))
        _create_file(os.path.join(self.snap_dir, "hooks", "test-hook"))

        # Now write the snap directory.
        self.generate_meta_yaml()

        # Ensure the file is executable
        self.assertThat(os.path.join(self.hooks_dir, "test-hook"), FileExists())
        self.assertTrue(
            os.stat(os.path.join(self.hooks_dir, "test-hook")).st_mode & stat.S_IEXEC
        )

    def test_snap_hooks_not_does_not_fail_on_symlink(self):
        # Setup a snap directory containing a few things.
        _create_file(os.path.join(self.snap_dir, "snapcraft.yaml"))
        _create_file(os.path.join(self.snap_dir, "hooks", "test-hook"))
        # Create a working symlink
        os.symlink(
            "test-hook", os.path.join(self.snap_dir, "hooks", "test-hook-symlink")
        )

        # Now write the snap directory.
        self.generate_meta_yaml()

        # Ensure the file is executable
        self.assertThat(os.path.join(self.hooks_dir, "test-hook"), FileExists())
        test_hook_stat = os.stat(
            os.path.join(self.hooks_dir, "test-hook"), follow_symlinks=False
        )
        test_hook_symlink_stat = os.stat(
            os.path.join(self.hooks_dir, "test-hook-symlink"), follow_symlinks=False
        )

        self.assertThat(test_hook_symlink_stat.st_ino, Equals(test_hook_stat.st_ino))


class GenerateHookWrappersTestCase(CreateBaseTestCase):
    def test_generate_hook_wrappers(self):
        # Set up the prime directory to contain a few hooks in snap/hooks
        snap_hooks_dir = os.path.join(self.prime_dir, "snap", "hooks")
        hook1_path = os.path.join(snap_hooks_dir, "test-hook1")
        hook2_path = os.path.join(snap_hooks_dir, "test-hook2")

        for path in (hook1_path, hook2_path):
            _create_file(path, executable=True)

        # Now generate hook wrappers, and verify that they're correct
        self.generate_meta_yaml()
        for hook in ("test-hook1", "test-hook2"):
            hook_path = os.path.join(self.hooks_dir, hook)
            self.assertThat(hook_path, FileExists())
            self.assertThat(hook_path, unit.IsExecutable())

            # The hook in meta/hooks should exec the one in snap/hooks, as it's
            # a wrapper generated by snapcraft.
            self.assertThat(
                hook_path,
                FileContains(
                    matcher=Contains('exec "$SNAP/snap/hooks/{}"'.format(hook))
                ),
            )

    @patch("snapcraft.internal.project_loader._config.Config.snap_env")
    def test_generated_hook_wrappers_include_environment(self, mock_snap_env):
        mock_snap_env.return_value = ["PATH={}/foo".format(self.prime_dir)]

        # Set up the prime directory to contain a hook in snap/hooks as well as
        # one in meta/hooks
        snap_hook = os.path.join(self.prime_dir, "snap", "hooks", "snap-hook")
        meta_hook = os.path.join(self.prime_dir, "meta", "hooks", "meta-hook")

        for path in (snap_hook, meta_hook):
            _create_file(path, executable=True, content=path)

        # Now generate hook wrappers
        self.generate_meta_yaml()

        # Verify that the hook already in meta was unchanged (no environment)
        final_meta_hook = os.path.join(self.hooks_dir, "meta-hook")
        self.assertThat(final_meta_hook, FileExists())
        self.assertThat(final_meta_hook, unit.IsExecutable())
        self.assertThat(final_meta_hook, FileContains(meta_hook))

        # Verify that the snap hook was unchanged
        self.assertThat(snap_hook, FileExists())
        self.assertThat(snap_hook, unit.IsExecutable())
        self.assertThat(snap_hook, FileContains(snap_hook))

        # Verify that the snap hook got a wrapper generated for it with a full
        # environment
        final_snap_hook = os.path.join(self.hooks_dir, "snap-hook")
        self.assertThat(final_snap_hook, FileExists())
        self.assertThat(final_snap_hook, unit.IsExecutable())
        expected = textwrap.dedent(
            """\
            #!/bin/sh
            export PATH=$SNAP/foo
            export LD_LIBRARY_PATH="$SNAP_LIBRARY_PATH${LD_LIBRARY_PATH:+:$LD_LIBRARY_PATH}"
            exec "$SNAP/snap/hooks/snap-hook" "$@"
        """
        )

        self.assertThat(final_snap_hook, FileContains(expected))

    def test_generate_hook_wrappers_not_executable_chmods(self):
        # Set up the prime directory to contain a hook in snap/hooks that is
        # not executable.
        snap_hooks_dir = os.path.join(self.prime_dir, "snap", "hooks")
        _create_file(os.path.join(snap_hooks_dir, "test-hook"))

        # Now generate hook wrappers.
        self.generate_meta_yaml()

        # Ensure the file is executable
        self.assertThat(os.path.join(self.hooks_dir, "test-hook"), FileExists())
        self.assertTrue(
            os.stat(os.path.join(self.hooks_dir, "test-hook")).st_mode & stat.S_IEXEC
        )


class TestConfinement(CreateBaseTestCase):
    def test_strict(self):
        self.config_data["confinement"] = "strict"

        self.assertThat(self.generate_meta_yaml()["confinement"], Equals("strict"))

    def test_devmode(self):
        self.config_data["confinement"] = "devmode"

        self.assertThat(self.generate_meta_yaml()["confinement"], Equals("devmode"))

    def test_no_confinement(self):
        fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(fake_logger)

        del self.config_data["confinement"]

        # Ensure confinement defaults to strict if not specified. Also
        # verify that a warning is printed
        self.assertThat(self.generate_meta_yaml()["confinement"], Equals("strict"))
        self.assertThat(
            fake_logger.output,
            Contains("'confinement' property not specified: defaulting to 'strict'"),
        )


class TestRootEnvironmentLibraryPathWarnings(CreateBaseTestCase):
    def setUp(self):
        super().setUp()

        self.config_data["grade"] = "stable"

        self.fake_logger = fixtures.FakeLogger(level=logging.WARNING)
        self.useFixture(self.fake_logger)

    def assert_warnings(self):
        self.generate_meta_yaml()

        self.assertThat(
            self.fake_logger.output,
            Contains(
                "CVE-2020-27348: A potentially empty LD_LIBRARY_PATH has been set for environment "
                "in '.'. The current working directory will be added to the library path if empty. "
                "This can cause unexpected libraries to be loaded."
            ),
        )

    def assert_no_warnings(self):
        self.generate_meta_yaml()

        self.assertThat(
            self.fake_logger.output,
            Not(
                Contains(
                    "CVE-2020-27348: A potentially empty LD_LIBRARY_PATH has been set for environment "
                    "in '.'. The current working directory will be added to the library path if empty. "
                    "This can cause unexpected libraries to be loaded."
                )
            ),
        )

    def test_root_ld_library_path(self):
        self.config_data["environment"] = {"LD_LIBRARY_PATH": "/foo:$LD_LIBRARY_PATH"}

        self.assert_warnings()

    def test_root_ld_library_path_braces(self):
        self.config_data["environment"] = {"LD_LIBRARY_PATH": "/foo:${LD_LIBRARY_PATH}"}

        self.assert_warnings()

    def test_root_ld_library_path_with_colon_at_start(self):
        self.config_data["environment"] = {"LD_LIBRARY_PATH": ":/foo:/bar"}

        self.assert_warnings()

    def test_root_ld_library_path_with_colon_at_end(self):
        self.config_data["environment"] = {"LD_LIBRARY_PATH": "/foo:/bar:"}

        self.assert_warnings()

    def test_root_ld_library_path_with_colon_at_middle(self):
        self.config_data["environment"] = {"LD_LIBRARY_PATH": "/foo::/bar"}

        self.assert_warnings()

    def test_root_ld_library_path_good(self):
        self.config_data["environment"] = {"LD_LIBRARY_PATH": "/foo:/bar"}

        self.assert_no_warnings()


class TestAppsEnvironmentLibraryPathWarnings(CreateBaseTestCase):
    def setUp(self):
        super().setUp()

        self.config_data["grade"] = "stable"
        self.config_data["apps"] = {
            "app1": {"command": "foo"},
            "app2": {"command": "bar"},
        }

        _create_file(os.path.join(self.prime_dir, "foo"), executable=True)
        _create_file(os.path.join(self.prime_dir, "bar"), executable=True)

        self.fake_logger = fixtures.FakeLogger(level=logging.WARNING)
        self.useFixture(self.fake_logger)

    def assert_warnings_both(self):
        self.generate_meta_yaml()

        self.assertThat(
            self.fake_logger.output,
            Contains(
                "CVE-2020-27348: A potentially empty LD_LIBRARY_PATH has been set for environment "
                "in 'app1' and 'app2'. The current working directory will be added to the library "
                "path if empty. "
                "This can cause unexpected libraries to be loaded."
            ),
        )

    def assert_warnings_app1(self):
        self.generate_meta_yaml()

        self.assertThat(
            self.fake_logger.output,
            Contains(
                "CVE-2020-27348: A potentially empty LD_LIBRARY_PATH has been set for environment "
                "in 'app1'. The current working directory will be added to the library "
                "path if empty. "
                "This can cause unexpected libraries to be loaded."
            ),
        )

    def assert_no_warnings(self):
        self.generate_meta_yaml()

        self.assertThat(
            self.fake_logger.output,
            Not(
                Contains(
                    "CVE-2020-27348: A potentially empty LD_LIBRARY_PATH has been set for environment "
                    "in 'app1' and 'app2'. The current working directory will be added to the library "
                    "path if empty. "
                    "This can cause unexpected libraries to be loaded."
                )
            ),
        )

    def test_app_ld_library_path_app1(self):
        self.config_data["apps"]["app1"]["environment"] = {
            "LD_LIBRARY_PATH": "/foo:$LD_LIBRARY_PATH"
        }

        self.assert_warnings_app1()

    def test_app_ld_library_path_app1_braces(self):
        self.config_data["apps"]["app1"]["environment"] = {
            "LD_LIBRARY_PATH": "/foo:${LD_LIBRARY_PATH}"
        }

        self.assert_warnings_app1()

    def test_app_ld_library_path_app1_app2(self):
        self.config_data["apps"]["app1"]["environment"] = {
            "LD_LIBRARY_PATH": "/foo:$LD_LIBRARY_PATH"
        }
        self.config_data["apps"]["app2"]["environment"] = {
            "LD_LIBRARY_PATH": "/foo:$LD_LIBRARY_PATH"
        }

        self.assert_warnings_both()

    def test_root_ld_library_path_set(self):
        self.config_data["environment"] = {"LD_LIBRARY_PATH": "/foo:/bar"}

        self.config_data["apps"]["app1"]["environment"] = {
            "LD_LIBRARY_PATH": "/foo:$LD_LIBRARY_PATH"
        }
        self.config_data["apps"]["app2"]["environment"] = {
            "LD_LIBRARY_PATH": "/foo:$LD_LIBRARY_PATH"
        }

        self.assert_no_warnings()

    def test_app_ld_library_path_colon_middle_app1_app2(self):
        self.config_data["apps"]["app1"]["environment"] = {
            "LD_LIBRARY_PATH": "/foo::/bar"
        }
        self.config_data["apps"]["app2"]["environment"] = {
            "LD_LIBRARY_PATH": "/foo::/bar"
        }

        self.assert_warnings_both()


class TestGrade(CreateBaseTestCase):
    def test_stable(self):
        self.config_data["grade"] = "stable"

        self.assertThat(self.generate_meta_yaml()["grade"], Equals("stable"))

    def test_devel(self):
        self.config_data["grade"] = "devel"

        self.assertThat(self.generate_meta_yaml()["grade"], Equals("devel"))

    def test_no_grade(self):
        fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(fake_logger)

        # Ensure grade defaults to strict if not specified. Also
        # verify that a warning is printed
        self.assertThat(self.generate_meta_yaml()["grade"], Equals("stable"))
        self.assertThat(
            fake_logger.output,
            Contains("'grade' property not specified: defaulting to 'stable'"),
        )


class RequiredGradeTest(CreateBaseTestCase):
    def test_defaults_from_schema(self):
        self.assertThat(self.generate_meta_yaml()["grade"], Equals("stable"))

    def test_stable_required(self):
        global_state_path = "global_state"
        self.useFixture(
            fixtures.MockPatch(
                "snapcraft.project.Project._get_global_state_file_path",
                return_value=global_state_path,
            )
        )

        global_state = states.GlobalState()
        global_state.set_required_grade("stable")
        global_state.save(filepath=global_state_path)

        self.config_data["grade"] = "stable"

        self.assertThat(self.generate_meta_yaml()["grade"], Equals("stable"))

    def test_stable_but_devel_required(self):
        global_state_path = "global_state"
        self.useFixture(
            fixtures.MockPatch(
                "snapcraft.project.Project._get_global_state_file_path",
                return_value=global_state_path,
            )
        )

        global_state = states.GlobalState()
        global_state.set_required_grade("devel")
        global_state.save(filepath=global_state_path)

        self.config_data["grade"] = "stable"

        self.assertRaises(meta_errors.GradeDevelRequiredError, self.generate_meta_yaml)


class CommonIdTestCase(CreateBaseTestCase):
    def setUp(self):
        super().setUp()

        self.fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(self.fake_logger)

        self.create_metadata_file("part", "1.metainfo.xml", "test.id.1")
        self.create_metadata_file("part", "2.metainfo.xml", "test.id.2")

    def create_metadata_file(self, part, name, common_id):
        # Create metadata files
        filename = os.path.join("parts", part, "src", name)
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        with open(filename, "w") as f:
            f.write(
                textwrap.dedent(
                    """\
                    <?xml version="1.0" encoding="utf-8"?>
                    <component type="desktop">
                      <id>{}</id>
                    </component>
                    """.format(
                        common_id
                    )
                )
            )

    def make_snapcraft_project(self, common_id):
        yaml = textwrap.dedent(
            """\
           name: test
           base: core18
           version: "1.0"
           summary: test
           description: test
           confinement: strict
           grade: stable
           adopt-info: part
           apps:
             test-app:
               command: echo
               common-id: {common_id}
           parts:
             part:
               plugin: nil
               parse-info: ["1.metainfo.xml", "2.metainfo.xml"]
           """
        )

        yaml_path = self.make_snapcraft_yaml(yaml.format(common_id=common_id))
        project = Project(snapcraft_yaml_file_path=yaml_path)
        return project_loader.load_config(project)

    def test_common_id(self):
        config = self.make_snapcraft_project(common_id="test.id.2")
        for part in config.parts.all_parts:
            part.makedirs()
            part.mark_pull_done()
            part.build()
            part.stage()
            part.prime()
        _snap_packaging.create_snap_packaging(config)
        self.assertNotIn(
            "is not used in any appstream metafile.", self.fake_logger.output
        )

    def test_common_id_mismatch(self):
        config = self.make_snapcraft_project(common_id="test.id.mismatch")
        for part in config.parts.all_parts:
            part.makedirs()
            part.mark_pull_done()
            part.build()
            part.stage()
            part.prime()
        _snap_packaging.create_snap_packaging(config)
        self.assertIn(
            "Common ID 'test.id.mismatch' specified in app 'test-app' is not used in any metadata file.",
            self.fake_logger.output.strip(),
        )


def _create_file(path, *, content="", executable=False):
    basepath = os.path.dirname(path)
    if basepath:
        os.makedirs(basepath, exist_ok=True)
    mode = "wb" if type(content) == bytes else "w"
    with open(path, mode) as f:
        f.write(content)
    if executable:
        os.chmod(path, 0o755)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

import fixtures
from testtools import TestCase
from testtools.matchers import Equals

from snapcraft.internal import lifecycle, project_loader, states, steps
from snapcraft.project import Project
from snapcraft.storeapi.errors import SnapNotFoundError
from snapcraft.storeapi.info import SnapInfo
from tests import fixture_setup


class TestGlobalState(TestCase):
    def setUp(self):
        super().setUp()

        self.useFixture(fixtures.FakeLogger(level=logging.ERROR))

        temp_cwd = fixture_setup.TempCWD()
        self.useFixture(temp_cwd)

        snapcraft_yaml = fixture_setup.SnapcraftYaml(
            temp_cwd.path, base="core18", parts={"test-part": {"plugin": "nil"}}
        )
        self.useFixture(snapcraft_yaml)

        project = Project(
            snapcraft_yaml_file_path=snapcraft_yaml.snapcraft_yaml_file_path
        )

        self.global_state_filepath = project._get_global_state_file_path()

        self.project_config = project_loader.load_config(project)

        self.useFixture(
            fixtures.MockPatchObject(
                self.project_config, "get_build_snaps", return_value={"core18"}
            )
        )

        self.useFixture(
            fixtures.MockPatch("snapcraft.internal.lifecycle._runner._Executor.run")
        )

        self.useFixture(
            fixtures.MockPatch("snapcraft.internal.repo.snaps.install_snaps")
        )

        # Avoid unnecessary calls to info.
        channel_map = []
        for arch in ("amd64", "i386", "s390x", "arm64", "armhf", "ppc64el"):
            channel_map.append(
                {
                    "channel": {
                        "architecture": arch,
                        "name": "stable",
                        "released-at": "2019-10-15T13:54:06.800280+00:00",
                        "risk": "stable",
                        "track": "latest",
                    },
                    "confinement": "strict",
                    "download": {
                        "deltas": [],
                        "sha3-384": "64d232d6bfa65be14d7f8d84e952d4e372e12021e2c3dbaf70cf2af5e78bf51c4baf9c9107dd6db815064636b781bda6",
                        "size": 57151488,
                        "url": "https://api.snapcraft.io/api/v1/snaps/download/CSO04Jhav2yK0uz97cr0ipQRyqg0qQL6_1223.snap",
                    },
                    "revision": 1223,
                }
            )
        info = {
            "channel-map": channel_map,
            "default-track": None,
            "name": "core18",
            "snap": {
                "name": "core18",
                "publisher": {
                    "display-name": "Canonical",
                    "id": "canonical",
                    "username": "canonical",
                    "validation": "verified",
                },
                "snap-id": "CSO04Jhav2yK0uz97cr0ipQRyqg0qQL6",
            },
            "snap-id": "CSO04Jhav2yK0uz97cr0ipQRyqg0qQL6",
        }
        self.fake_storeapi_get_info = fixtures.MockPatch(
            "snapcraft.storeapi._snap_api.SnapAPI.get_info",
            return_value=SnapInfo(info),
        )
        self.useFixture(self.fake_storeapi_get_info)

    def test_stable_grade_for_stable_base(self):
        lifecycle.execute(steps.PULL, self.project_config)

        global_state = states.GlobalState.load(filepath=self.global_state_filepath)
        self.assertThat(global_state.get_required_grade(), Equals("stable"))
        self.fake_storeapi_get_info.mock.assert_called_once_with("core18")

    def test_stable_grade_for_non_stable_base(self):
        self.fake_storeapi_get_info.mock.side_effect = SnapNotFoundError(
            snap_name="core18"
        )
        lifecycle.execute(steps.PULL, self.project_config)

        global_state = states.GlobalState.load(filepath=self.global_state_filepath)
        self.assertThat(global_state.get_required_grade(), Equals("devel"))
        self.fake_storeapi_get_info.mock.assert_called_once_with("core18")

    def test_grade_not_queried_for_if_already_set(self):
        # Set the grade
        global_state = states.GlobalState()
        global_state.set_required_grade("devel")
        global_state.save(filepath=self.global_state_filepath)

        lifecycle.execute(steps.PULL, self.project_config)

        self.fake_storeapi_get_info.mock.assert_not_called()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The meson plugin is useful for building meson based parts.

Meson based projects are projects that have a meson.build that drives the
build.

This plugin leverages ninja to build and install.

Additionally, this plugin uses the following plugin-specific keywords:

    - meson-version
      (string)
      The version of meson to install from PyPI.
      If unspecified, the latest released version of meson will be used.

    - meson-parameters
      (list of strings)
      Configure flags to pass to the build using the common meson semantics.
"""

from typing import Any, Dict, List, Set

from snapcraft.plugins.v2 import PluginV2


class MesonPlugin(PluginV2):
    @classmethod
    def get_schema(cls) -> Dict[str, Any]:
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "meson-parameters": {
                    "type": "array",
                    "uniqueItems": True,
                    "items": {"type": "string"},
                    "default": [],
                },
                "meson-version": {"type": "string", "default": ""},
            },
            "required": ["source"],
        }

    def get_build_snaps(self) -> Set[str]:
        return set()

    def get_build_packages(self) -> Set[str]:
        return {
            "ninja-build",
            "gcc",
            "python3-pip",
            "python3-setuptools",
            "python3-wheel",
        }

    def get_build_environment(self) -> Dict[str, str]:
        return dict()

    @property
    def out_of_source_build(self):
        return True

    def get_build_commands(self) -> List[str]:
        if self.options.meson_version:
            meson_package = f"meson=={self.options.meson_version}"
        else:
            meson_package = "meson"

        meson_cmd = ["meson"]
        if self.options.meson_parameters:
            meson_cmd.append(" ".join(self.options.meson_parameters))
        meson_cmd.append('"${SNAPCRAFT_PART_SRC_WORK}"')

        return [
            f"/usr/bin/python3 -m pip install -U {meson_package}",
            "[ ! -f build.ninja ] && {}".format(" ".join(meson_cmd)),
            "ninja",
            'DESTDIR="${SNAPCRAFT_PART_INSTALL}" ninja install',
        ]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import shutil
import subprocess
from unittest import mock

import fixtures
from testtools.matchers import Equals

from snapcraft.internal import sources
from snapcraft.internal.sources import errors
from tests import unit
from tests.subprocess_utils import call, call_with_output


def fake_git_command_error(*args, **kwargs):
    raise subprocess.CalledProcessError(44, ["git"], output=b"git: some error")


# LP: #1733584
class TestGit(unit.sources.SourceTestCase):  # type: ignore
    def setUp(self):

        super().setUp()
        patcher = mock.patch("snapcraft.sources.Git._get_source_details")
        self.mock_get_source_details = patcher.start()
        self.mock_get_source_details.return_value = ""
        self.addCleanup(patcher.stop)

        self.fake_check_output = self.useFixture(
            fixtures.MockPatch("subprocess.check_output")
        ).mock

    def test_pull(self):
        git = sources.Git("git://my-source", "source_dir")

        git.pull()

        self.mock_run.assert_called_once_with(
            ["git", "clone", "--recursive", "git://my-source", "source_dir"]
        )

    def test_add(self):
        url = "git://my-source"
        source_dir = "source_dir"

        git = sources.Git(url, source_dir)
        git.add("file")
        self.fake_check_output.assert_called_once_with(
            ["git", "-C", "source_dir", "add", "file"]
        )

    def test_add_error(self):
        self.fake_check_output.side_effect = fake_git_command_error
        url = "git://my-source"
        source_dir = "source_dir"

        git = sources.Git(url, source_dir)

        error = self.assertRaises(errors.GitCommandError, git.add, "file")
        self.assertThat(
            error.get_brief(),
            Equals("Failed to execute git command: git -C source_dir add file"),
        )

    def test_add_abs_path(self):
        url = "git://my-source"
        source_dir = "source_dir"

        git = sources.Git(url, source_dir)
        git.add(os.path.join(source_dir, "file"))
        self.fake_check_output.assert_called_once_with(
            ["git", "-C", "source_dir", "add", "file"]
        )

    def test_commit(self):
        url = "git://my-source"
        source_dir = "source_dir"

        git = sources.Git(url, source_dir)
        git.commit(message="message", author="author")

        self.fake_check_output.assert_called_once_with(
            [
                "git",
                "-C",
                "source_dir",
                "commit",
                "--no-gpg-sign",
                "--message",
                "message",
                "--author",
                "author",
            ]
        )

    def test_commit_error(self):
        self.fake_check_output.side_effect = fake_git_command_error
        url = "git://my-source"
        source_dir = "source_dir"

        git = sources.Git(url, source_dir)

        error = self.assertRaises(
            errors.GitCommandError, git.commit, message="message", author="author"
        )
        self.assertThat(
            error.get_brief(),
            Equals(
                "Failed to execute git command: git -C source_dir commit --no-gpg-sign --message message --author author"
            ),
        )

    def test_init(self):
        url = "git://my-source"
        source_dir = "source_dir"

        git = sources.Git(url, source_dir)
        git.init()
        self.fake_check_output.assert_called_once_with(
            ["git", "-C", "source_dir", "init"]
        )

    def test_init_error(self):
        self.fake_check_output.side_effect = fake_git_command_error
        url = "git://my-source"
        source_dir = "source_dir"

        git = sources.Git(url, source_dir)

        error = self.assertRaises(errors.GitCommandError, git.init)
        self.assertThat(
            error.get_brief(),
            Equals("Failed to execute git command: git -C source_dir init"),
        )

    def test_push(self):
        url = "git://my-source"
        refspec = "HEAD:master"
        source_dir = "source_dir"

        git = sources.Git(url, source_dir)
        git.push(url, refspec)
        self.fake_check_output.assert_called_once_with(
            ["git", "-C", "source_dir", "push", url, refspec]
        )

    def test_push_force(self):
        url = "git://my-source"
        refspec = "HEAD:master"
        source_dir = "source_dir"

        git = sources.Git(url, source_dir)
        git.push(url, refspec, force=True)
        self.fake_check_output.assert_called_once_with(
            ["git", "-C", "source_dir", "push", url, refspec, "--force"]
        )

    def test_push_error(self):
        self.fake_check_output.side_effect = fake_git_command_error
        url = "git://my-source"
        refspec = "HEAD:master"
        source_dir = "source_dir"

        git = sources.Git(url, source_dir)
        error = self.assertRaises(errors.GitCommandError, git.push, url, refspec)
        self.assertThat(
            error.get_brief(),
            Equals(
                "Failed to execute git command: git -C source_dir push git://my-source HEAD:master"
            ),
        )

    def test_pull_with_depth(self):
        git = sources.Git("git://my-source", "source_dir", source_depth=2)

        git.pull()

        self.mock_run.assert_called_once_with(
            [
                "git",
                "clone",
                "--recursive",
                "--depth",
                "2",
                "git://my-source",
                "source_dir",
            ]
        )

    def test_pull_branch(self):
        git = sources.Git("git://my-source", "source_dir", source_branch="my-branch")
        git.pull()

        self.mock_run.assert_called_once_with(
            [
                "git",
                "clone",
                "--recursive",
                "--branch",
                "my-branch",
                "git://my-source",
                "source_dir",
            ]
        )

    def test_pull_tag(self):
        git = sources.Git("git://my-source", "source_dir", source_tag="tag")
        git.pull()

        self.mock_run.assert_called_once_with(
            [
                "git",
                "clone",
                "--recursive",
                "--branch",
                "tag",
                "git://my-source",
                "source_dir",
            ]
        )

    def test_pull_commit(self):
        git = sources.Git(
            "git://my-source",
            "source_dir",
            source_commit="2514f9533ec9b45d07883e10a561b248497a8e3c",
        )
        git.pull()

        self.mock_run.assert_has_calls(
            [
                mock.call(
                    ["git", "clone", "--recursive", "git://my-source", "source_dir"]
                ),
                mock.call(
                    [
                        "git",
                        "-C",
                        "source_dir",
                        "fetch",
                        "origin",
                        "2514f9533ec9b45d07883e10a561b248497a8e3c",
                    ]
                ),
                mock.call(
                    [
                        "git",
                        "-C",
                        "source_dir",
                        "checkout",
                        "2514f9533ec9b45d07883e10a561b248497a8e3c",
                    ]
                ),
            ]
        )

    def test_pull_existing(self):
        self.mock_path_exists.return_value = True

        git = sources.Git("git://my-source", "source_dir")
        git.pull()

        self.mock_run.assert_has_calls(
            [
                mock.call(
                    [
                        "git",
                        "-C",
                        "source_dir",
                        "fetch",
                        "--prune",
                        "--recurse-submodules=yes",
                    ]
                ),
                mock.call(
                    ["git", "-C", "source_dir", "reset", "--hard", "origin/master"]
                ),
                mock.call(
                    [
                        "git",
                        "-C",
                        "source_dir",
                        "submodule",
                        "update",
                        "--recursive",
                        "--force",
                    ]
                ),
            ]
        )

    def test_pull_existing_with_tag(self):
        self.mock_path_exists.return_value = True

        git = sources.Git("git://my-source", "source_dir", source_tag="tag")
        git.pull()

        self.mock_run.assert_has_calls(
            [
                mock.call(
                    [
                        "git",
                        "-C",
                        "source_dir",
                        "fetch",
                        "--prune",
                        "--recurse-submodules=yes",
                    ]
                ),
                mock.call(
                    ["git", "-C", "source_dir", "reset", "--hard", "refs/tags/tag"]
                ),
                mock.call(
                    [
                        "git",
                        "-C",
                        "source_dir",
                        "submodule",
                        "update",
                        "--recursive",
                        "--force",
                    ]
                ),
            ]
        )

    def test_pull_existing_with_commit(self):
        self.mock_path_exists.return_value = True

        git = sources.Git(
            "git://my-source",
            "source_dir",
            source_commit="2514f9533ec9b45d07883e10a561b248497a8e3c",
        )
        git.pull()

        self.mock_run.assert_has_calls(
            [
                mock.call(
                    [
                        "git",
                        "-C",
                        "source_dir",
                        "fetch",
                        "--prune",
                        "--recurse-submodules=yes",
                    ]
                ),
                mock.call(
                    [
                        "git",
                        "-C",
                        "source_dir",
                        "reset",
                        "--hard",
                        "2514f9533ec9b45d07883e10a561b248497a8e3c",
                    ]
                ),
                mock.call(
                    [
                        "git",
                        "-C",
                        "source_dir",
                        "submodule",
                        "update",
                        "--recursive",
                        "--force",
                    ]
                ),
            ]
        )

    def test_pull_existing_with_branch(self):
        self.mock_path_exists.return_value = True

        git = sources.Git("git://my-source", "source_dir", source_branch="my-branch")
        git.pull()

        self.mock_run.assert_has_calls(
            [
                mock.call(
                    [
                        "git",
                        "-C",
                        "source_dir",
                        "fetch",
                        "--prune",
                        "--recurse-submodules=yes",
                    ]
                ),
                mock.call(
                    [
                        "git",
                        "-C",
                        "source_dir",
                        "reset",
                        "--hard",
                        "refs/heads/my-branch",
                    ]
                ),
                mock.call(
                    [
                        "git",
                        "-C",
                        "source_dir",
                        "submodule",
                        "update",
                        "--recursive",
                        "--force",
                    ]
                ),
            ]
        )

    def test_init_with_source_branch_and_tag_raises_exception(self):
        raised = self.assertRaises(
            sources.errors.SnapcraftSourceIncompatibleOptionsError,
            sources.Git,
            "git://mysource",
            "source_dir",
            source_tag="tag",
            source_branch="branch",
        )

        self.assertThat(raised.source_type, Equals("git"))
        self.assertThat(raised.options, Equals(["source-tag", "source-branch"]))

    def test_init_with_source_branch_and_commit_raises_exception(self):
        raised = self.assertRaises(
            sources.errors.SnapcraftSourceIncompatibleOptionsError,
            sources.Git,
            "git://mysource",
            "source_dir",
            source_commit="2514f9533ec9b45d07883e10a561b248497a8e3c",
            source_branch="branch",
        )

        self.assertThat(raised.source_type, Equals("git"))
        self.assertThat(raised.options, Equals(["source-branch", "source-commit"]))

    def test_init_with_source_tag_and_commit_raises_exception(self):
        raised = self.assertRaises(
            sources.errors.SnapcraftSourceIncompatibleOptionsError,
            sources.Git,
            "git://mysource",
            "source_dir",
            source_commit="2514f9533ec9b45d07883e10a561b248497a8e3c",
            source_tag="tag",
        )

        self.assertThat(raised.source_type, Equals("git"))
        self.assertThat(raised.options, Equals(["source-tag", "source-commit"]))

    def test_source_checksum_raises_exception(self):
        raised = self.assertRaises(
            sources.errors.SnapcraftSourceInvalidOptionError,
            sources.Git,
            "git://mysource",
            "source_dir",
            source_checksum="md5/d9210476aac5f367b14e513bdefdee08",
        )

        self.assertThat(raised.source_type, Equals("git"))
        self.assertThat(raised.option, Equals("source-checksum"))

    def test_has_source_handler_entry(self):
        self.assertTrue(sources._source_handler["git"] is sources.Git)

    def test_pull_failure(self):
        self.mock_run.side_effect = subprocess.CalledProcessError(1, [])

        git = sources.Git("git://my-source", "source_dir")
        raised = self.assertRaises(sources.errors.SnapcraftPullError, git.pull)
        self.assertThat(
            raised.command, Equals("git clone --recursive git://my-source source_dir")
        )
        self.assertThat(raised.exit_code, Equals(1))


class GitBaseTestCase(unit.TestCase):
    def rm_dir(self, dir):
        if os.path.exists(dir):
            shutil.rmtree(dir)

    def clean_dir(self, dir):
        self.rm_dir(dir)
        os.mkdir(dir)
        self.addCleanup(self.rm_dir, dir)

    def clone_repo(self, repo, tree):
        self.clean_dir(tree)
        call(["git", "clone", repo, tree])
        os.chdir(tree)
        call(["git", "config", "--local", "user.name", '"Example Dev"'])
        call(["git", "config", "--local", "user.email", "dev@example.com"])

    def add_file(self, filename, body, message):
        with open(filename, "w") as fp:
            fp.write(body)

        call(["git", "add", filename])
        call(["git", "commit", "-am", message])

    def check_file_contents(self, path, expected):
        body = None
        with open(path) as fp:
            body = fp.read()
        self.assertThat(body, Equals(expected))


class TestGitConflicts(GitBaseTestCase):
    """Test that git pull errors don't kill the parser"""

    def test_git_conflicts(self):

        repo = "/tmp/conflict-test.git"
        working_tree = "/tmp/git-conflict-test"
        conflicting_tree = "{}-conflict".format(working_tree)
        git = sources.Git(repo, working_tree, silent=True)

        self.clean_dir(repo)
        self.clean_dir(working_tree)
        self.clean_dir(conflicting_tree)

        os.chdir(repo)
        call(["git", "init", "--bare"])

        self.clone_repo(repo, working_tree)

        # check out the original repo
        self.clone_repo(repo, conflicting_tree)

        # add a file to the repo
        os.chdir(working_tree)
        self.add_file("fake", "fake 1", "fake 1")
        call(["git", "push", repo])

        git.pull()

        os.chdir(conflicting_tree)
        self.add_file("fake", "fake 2", "fake 2")
        call(["git", "push", "-f", repo])

        os.chdir(working_tree)
        git.pull()

        body = None
        with open(os.path.join(working_tree, "fake")) as fp:
            body = fp.read()

        self.assertThat(body, Equals("fake 2"))

    def test_git_submodules(self):
        """Test that updates to submodules are pulled"""
        repo = "/tmp/submodules.git"
        sub_repo = "/tmp/subrepo"
        working_tree = "/tmp/git-submodules"
        working_tree_two = "{}-two".format(working_tree)
        sub_working_tree = "/tmp/git-submodules-sub"
        git = sources.Git(repo, working_tree, silent=True)

        self.clean_dir(repo)
        self.clean_dir(sub_repo)
        self.clean_dir(working_tree)
        self.clean_dir(working_tree_two)
        self.clean_dir(sub_working_tree)

        os.chdir(sub_repo)
        call(["git", "init", "--bare"])

        self.clone_repo(sub_repo, sub_working_tree)
        self.add_file("sub-file", "sub-file", "sub-file")
        call(["git", "push", sub_repo])

        os.chdir(repo)
        call(["git", "init", "--bare"])

        self.clone_repo(repo, working_tree)
        call(["git", "submodule", "add", sub_repo])
        call(["git", "commit", "-am", "added submodule"])
        call(["git", "push", repo])

        git.pull()

        self.check_file_contents(
            os.path.join(working_tree, "subrepo", "sub-file"), "sub-file"
        )

        # add a file to the repo
        os.chdir(sub_working_tree)
        self.add_file("fake", "fake 1", "fake 1")
        call(["git", "push", sub_repo])

        os.chdir(working_tree)
        git.pull()

        # this shouldn't cause any change
        self.check_file_contents(
            os.path.join(working_tree, "subrepo", "sub-file"), "sub-file"
        )
        self.assertFalse(os.path.exists(os.path.join(working_tree, "subrepo", "fake")))

        # update the submodule
        self.clone_repo(repo, working_tree_two)
        call(["git", "submodule", "update", "--init", "--recursive", "--remote"])
        call(["git", "add", "subrepo"])
        call(["git", "commit", "-am", "updated submodule"])
        call(["git", "push"])

        os.chdir(working_tree)
        git.pull()

        # new file should be there now
        self.check_file_contents(
            os.path.join(working_tree, "subrepo", "sub-file"), "sub-file"
        )
        self.check_file_contents(
            os.path.join(working_tree, "subrepo", "fake"), "fake 1"
        )


class GitDetailsTestCase(GitBaseTestCase):
    def setUp(self):
        def _add_and_commit_file(filename, content=None, message=None):
            if not content:
                content = filename
            if not message:
                message = filename

            with open(filename, "w") as fp:
                fp.write(content)
            call(["git", "add", filename])
            call(["git", "commit", "-am", message])

        super().setUp()
        self.working_tree = "git-test"
        self.source_dir = "git-checkout"
        self.clean_dir(self.working_tree)
        os.chdir(self.working_tree)
        call(["git", "init"])
        call(["git", "config", "user.name", '"Example Dev"'])
        call(["git", "config", "user.email", "dev@example.com"])
        _add_and_commit_file("testing")
        self.expected_commit = call_with_output(["git", "rev-parse", "HEAD"])

        _add_and_commit_file("testing-2")
        call(["git", "tag", "test-tag"])
        self.expected_tag = "test-tag"

        _add_and_commit_file("testing-3")
        self.expected_branch = "test-branch"
        call(["git", "branch", self.expected_branch])

        os.chdir("..")

        self.git = sources.Git(
            self.working_tree,
            self.source_dir,
            silent=True,
            source_commit=self.expected_commit,
        )
        self.git.pull()

        self.source_details = self.git._get_source_details()

    def test_git_details_commit(self):
        self.assertThat(
            self.source_details["source-commit"], Equals(self.expected_commit)
        )

    def test_git_details_branch(self):
        shutil.rmtree(self.source_dir)
        self.git = sources.Git(
            self.working_tree,
            self.source_dir,
            silent=True,
            source_branch=self.expected_branch,
        )
        self.git.pull()

        self.source_details = self.git._get_source_details()
        self.assertThat(
            self.source_details["source-branch"], Equals(self.expected_branch)
        )

    def test_git_details_tag(self):
        self.git = sources.Git(
            self.working_tree, self.source_dir, silent=True, source_tag="test-tag"
        )
        self.git.pull()

        self.source_details = self.git._get_source_details()
        self.assertThat(self.source_details["source-tag"], Equals(self.expected_tag))


class GitGenerateVersionBaseTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()

        patcher = mock.patch("subprocess.check_output")
        self.output_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("subprocess.Popen")
        self.popen_mock = patcher.start()
        self.addCleanup(patcher.stop)


class TestGitGenerateVersion:

    scenarios = (
        ("only_tag", dict(return_value="2.28", expected="2.28")),
        (
            "tag+commits",
            dict(return_value="2.28-28-gabcdef1", expected="2.28+git28.abcdef1"),
        ),
        (
            "tag+dirty",
            dict(
                return_value="2.28-29-gabcdef1-dirty",
                expected="2.28+git29.abcdef1-dirty",
            ),
        ),
    )

    def test_version(self, monkeypatch, return_value, expected):
        monkeypatch.setattr(
            subprocess, "check_output", lambda *args, **kwargs: return_value.encode()
        )

        assert sources.Git.generate_version() == expected


class GitGenerateVersionNoTagTestCase(GitGenerateVersionBaseTestCase):
    def test_version(self):
        self.output_mock.side_effect = subprocess.CalledProcessError(1, [])
        proc_mock = mock.Mock()
        proc_mock.returncode = 0
        proc_mock.communicate.return_value = (b"abcdef1", b"")
        self.popen_mock.return_value = proc_mock

        expected = "0+git.abcdef1"
        self.assertThat(sources.Git.generate_version(), Equals(expected))


class GitGenerateVersionNoGitTestCase(GitGenerateVersionBaseTestCase):
    def test_version(self):
        self.output_mock.side_effect = subprocess.CalledProcessError(1, [])
        proc_mock = mock.Mock()
        proc_mock.returncode = 2
        proc_mock.communicate.return_value = (b"", b"No .git")
        self.popen_mock.return_value = proc_mock

        self.assertRaises(sources.errors.VCSError, sources.Git.generate_version)



import rclpy
from rcl_interfaces.msg import ParameterDescriptor
from rclpy.node import Node
from rclpy.qos import QoSProfile
from std_msgs.msg import String


class Listener(Node):
    def __init__(self):
        super().__init__("listener")

        parameter = self.declare_parameter(
            "exit_after_receive", False, ParameterDescriptor()
        )
        self._exit_after_receive = parameter.value
        self.should_exit = False

        qos_profile = QoSProfile(depth=1)
        self._subscription = self.create_subscription(
            String, "chatter", self._callback, qos_profile
        )

    def _callback(self, message):
        self.get_logger().info("I heard {!r}".format(message.data))
        if self._exit_after_receive:
            self.get_logger().info(
                "Requested to exit after message received. Exiting now."
            )
            self.should_exit = True


def main(args=None):
    rclpy.init(args=args)

    listener = Listener()

    while rclpy.ok():
        rclpy.spin_once(listener)
        if listener.should_exit:
            break

    listener.destroy_node()
    rclpy.shutdown()


if __name__ == "__main__":
    main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
import subprocess

from testtools.matchers import MatchesRegex

from tests import fixture_setup, integration, skip


class CatkinToolsTestCase(integration.SnapdIntegrationTestCase):

    slow_test = True

    @skip.skip_unless_codename("xenial", "ROS Kinetic only targets Xenial")
    def test_install_and_execution(self) -> None:
        self.useFixture(fixture_setup.WithoutSnapInstalled("catkin-tools-example"))
        self.run_snapcraft(project_dir="catkin-tools-talker-listener")
        self.install_snap()

        # Run the ROS system. By default this will never exit, but the code
        # supports an `exit-after-receive` parameter that, if true, will cause
        # the system to shutdown after the listener has successfully received
        # a message.
        output = subprocess.check_output(
            ["catkin-tools-example.launch-project", "exit-after-receive:=true"]
        ).decode()
        self.assertThat(
            output, MatchesRegex(r".*I heard Hello world.*", flags=re.DOTALL)
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest
from testtools.matchers import Equals, HasLength, Is, IsInstance

from snapcraft.storeapi.v2 import channel_map
from tests import unit


class ProgressiveTest(unit.TestCase):
    def test_progressive(self):
        payload = {"paused": False, "percentage": 83.3, "current-percentage": 32.1}

        p = channel_map.Progressive.unmarshal(payload)

        self.expectThat(repr(p), Equals(f"<Progressive: 32.1=>83.3>"))
        self.expectThat(p.paused, Equals(payload["paused"]))
        self.expectThat(p.percentage, Equals(payload["percentage"]))
        self.expectThat(p.current_percentage, Equals(payload["current-percentage"]))
        self.expectThat(p.marshal(), Equals(payload))

    def test_none(self):
        payload = {"paused": None, "percentage": None, "current-percentage": None}

        p = channel_map.Progressive.unmarshal(payload)

        self.expectThat(repr(p), Equals(f"<Progressive: None=>None>"))
        self.expectThat(p.paused, Equals(payload["paused"]))
        self.expectThat(p.percentage, Equals(payload["percentage"]))
        self.expectThat(p.current_percentage, Equals(payload["current-percentage"]))
        self.expectThat(p.marshal(), Equals(payload))


class MappedChannelTest(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.payload = {
            "architecture": "amd64",
            "channel": "latest/stable",
            "expiration-date": None,
            "progressive": {
                "paused": None,
                "percentage": None,
                "current-percentage": None,
            },
            "revision": 2,
        }

    def test_channel(self):
        mc = channel_map.MappedChannel.unmarshal(self.payload)

        self.expectThat(
            repr(mc),
            Equals(
                "<MappedChannel: 'latest/stable' for revision 2 and architecture 'amd64'>"
            ),
        )
        self.expectThat(mc.channel, Equals(self.payload["channel"]))
        self.expectThat(mc.revision, Equals(self.payload["revision"]))
        self.expectThat(mc.architecture, Equals(self.payload["architecture"]))
        self.expectThat(mc.progressive, IsInstance(channel_map.Progressive))
        self.expectThat(mc.expiration_date, Is(None))
        self.expectThat(mc.marshal(), Equals(self.payload))

    def test_channel_with_expiration(self):
        date_string = "2020-02-11T17:51:40.891996Z"
        self.payload.update({"expiration-date": date_string})

        mc = channel_map.MappedChannel.unmarshal(self.payload)

        self.expectThat(
            repr(mc),
            Equals(
                "<MappedChannel: 'latest/stable' for revision 2 and architecture 'amd64'>"
            ),
        )
        self.expectThat(mc.channel, Equals(self.payload["channel"]))
        self.expectThat(mc.revision, Equals(self.payload["revision"]))
        self.expectThat(mc.architecture, Equals(self.payload["architecture"]))
        self.expectThat(mc.progressive, IsInstance(channel_map.Progressive))
        self.expectThat(mc.expiration_date, Equals(date_string))
        self.expectThat(mc.marshal(), Equals(self.payload))


class SnapChannelTest(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.payload = {
            "name": "latest/candidate",
            "track": "latest",
            "risk": "candidate",
            "branch": None,
            "fallback": None,
        }

    def test_channel(self):
        sc = channel_map.SnapChannel.unmarshal(self.payload)

        self.expectThat(repr(sc), Equals("<SnapChannel: 'latest/candidate'>"))
        self.expectThat(sc.name, Equals(self.payload["name"]))
        self.expectThat(sc.track, Equals(self.payload["track"]))
        self.expectThat(sc.risk, Equals(self.payload["risk"]))
        self.expectThat(sc.branch, Is(None))
        self.expectThat(sc.fallback, Is(None))
        self.expectThat(sc.marshal(), Equals(self.payload))

    def test_channel_with_branch(self):
        self.payload.update({"branch": "test-branch"})

        sc = channel_map.SnapChannel.unmarshal(self.payload)

        self.expectThat(repr(sc), Equals("<SnapChannel: 'latest/candidate'>"))
        self.expectThat(sc.name, Equals(self.payload["name"]))
        self.expectThat(sc.track, Equals(self.payload["track"]))
        self.expectThat(sc.risk, Equals(self.payload["risk"]))
        self.expectThat(sc.branch, Equals(self.payload["branch"]))
        self.expectThat(sc.fallback, Is(None))
        self.expectThat(sc.marshal(), Equals(self.payload))

    def test_channel_with_fallback(self):
        self.payload.update({"fallback": "latest/stable"})

        sc = channel_map.SnapChannel.unmarshal(self.payload)

        self.expectThat(repr(sc), Equals("<SnapChannel: 'latest/candidate'>"))
        self.expectThat(sc.name, Equals(self.payload["name"]))
        self.expectThat(sc.track, Equals(self.payload["track"]))
        self.expectThat(sc.risk, Equals(self.payload["risk"]))
        self.expectThat(sc.branch, Is(None)),
        self.expectThat(sc.fallback, Equals(self.payload["fallback"]))
        self.expectThat(sc.marshal(), Equals(self.payload))


_TRACK_PAYLOADS = [
    {
        "name": "latest",
        "status": "active",
        "creation-date": None,
        "version-pattern": None,
    },
    {
        "name": "1.0",
        "status": "default",
        "creation-date": "2019-10-17T14:11:59Z",
        "version-pattern": "1.*",
    },
]


@pytest.mark.parametrize("payload", _TRACK_PAYLOADS)
def test_snap_track(payload):
    st = channel_map.SnapTrack.unmarshal(payload)

    assert repr(st) == f"<SnapTrack: {st.name!r}>"
    assert st.name == payload["name"]
    assert st.status == payload["status"]
    assert st.creation_date == payload["creation-date"]
    assert st.version_pattern == payload["version-pattern"]
    assert st.marshal() == payload


class RevisionTest(unit.TestCase):
    def test_revision(self):
        payload = {"revision": 2, "version": "2.0", "architectures": ["amd64", "arm64"]}

        r = channel_map.Revision.unmarshal(payload)

        self.expectThat(
            repr(r),
            Equals(
                "<Revision: 2 for version '2.0' and architectures ['amd64', 'arm64']>"
            ),
        )
        self.expectThat(r.revision, Equals(payload["revision"]))
        self.expectThat(r.version, Equals(payload["version"]))
        self.expectThat(r.architectures, Equals(payload["architectures"]))
        self.expectThat(r.marshal(), Equals(payload))


class SnapTest(unit.TestCase):
    def test_snap(self):
        payload = {
            "name": "my-snap",
            "channels": [
                {
                    "name": "latest/stable",
                    "track": "latest",
                    "risk": "candidate",
                    "branch": None,
                    "fallback": None,
                },
                {
                    "name": "latest/candidate",
                    "track": "latest",
                    "risk": "candidate",
                    "branch": None,
                    "fallback": "latest/stable",
                },
            ],
            "tracks": [
                {
                    "name": "track1",
                    "creation-date": "2019-10-17T14:11:59Z",
                    "status": "default",
                    "version-pattern": None,
                },
                {
                    "name": "track2",
                    "creation-date": None,
                    "status": "active",
                    "version-pattern": None,
                },
            ],
        }

        s = channel_map.Snap.unmarshal(payload)

        self.expectThat(repr(s), Equals("<Snap: 'my-snap'>"))
        self.expectThat(s.name, Equals(payload["name"]))

        snap_channels = s.channels
        self.expectThat(snap_channels, HasLength(2))
        self.expectThat(snap_channels[0], IsInstance(channel_map.SnapChannel))
        self.expectThat(snap_channels[1], IsInstance(channel_map.SnapChannel))

        self.expectThat(s.marshal(), Equals(payload))


class ChannelMapTest(unit.TestCase):
    def test_channel_map(self):
        payload = {
            "channel-map": [
                {
                    "architecture": "amd64",
                    "channel": "latest/stable",
                    "expiration-date": None,
                    "progressive": {
                        "paused": None,
                        "percentage": None,
                        "current-percentage": None,
                    },
                    "revision": 2,
                },
                {
                    "architecture": "amd64",
                    "channel": "latest/stable",
                    "expiration-date": None,
                    "progressive": {
                        "paused": None,
                        "percentage": 33.3,
                        "current-percentage": 12.3,
                    },
                    "revision": 3,
                },
                {
                    "architecture": "arm64",
                    "channel": "latest/stable",
                    "expiration-date": None,
                    "progressive": {
                        "paused": None,
                        "percentage": None,
                        "current-percentage": None,
                    },
                    "revision": 2,
                },
                {
                    "architecture": "i386",
                    "channel": "latest/stable",
                    "expiration-date": None,
                    "progressive": {
                        "paused": None,
                        "percentage": None,
                        "current-percentage": None,
                    },
                    "revision": 4,
                },
            ],
            "revisions": [
                {"revision": 2, "version": "2.0", "architectures": ["amd64", "arm64"]},
                {"revision": 3, "version": "2.0", "architectures": ["amd64", "arm64"]},
                {"revision": 4, "version": "2.0", "architectures": ["i386"]},
            ],
            "snap": {
                "name": "my-snap",
                "channels": [
                    {
                        "name": "latest/stable",
                        "track": "latest",
                        "risk": "candidate",
                        "branch": None,
                        "fallback": None,
                    },
                    {
                        "name": "latest/candidate",
                        "track": "latest",
                        "risk": "candidate",
                        "branch": None,
                        "fallback": "latest/stable",
                    },
                ],
                "tracks": [
                    {
                        "name": "track1",
                        "creation-date": "2019-10-17T14:11:59Z",
                        "status": "default",
                        "version-pattern": None,
                    },
                    {
                        "name": "track2",
                        "creation-date": None,
                        "status": "active",
                        "version-pattern": None,
                    },
                ],
            },
        }

        cm = channel_map.ChannelMap.unmarshal(payload)

        # Check "channel-map".
        self.expectThat(cm.channel_map, HasLength(4))
        self.expectThat(cm.channel_map[0], IsInstance(channel_map.MappedChannel))
        self.expectThat(cm.channel_map[1], IsInstance(channel_map.MappedChannel))
        self.expectThat(cm.channel_map[2], IsInstance(channel_map.MappedChannel))
        self.expectThat(cm.channel_map[3], IsInstance(channel_map.MappedChannel))

        # Check "revisions".
        self.expectThat(cm.revisions, HasLength(3))
        self.expectThat(cm.revisions[0], IsInstance(channel_map.Revision))
        self.expectThat(cm.revisions[1], IsInstance(channel_map.Revision))
        self.expectThat(cm.revisions[2], IsInstance(channel_map.Revision))

        # Check "snap".
        self.expectThat(cm.snap, IsInstance(channel_map.Snap))

        # Marshal.
        self.expectThat(cm.marshal(), Equals(payload))

        # Test the get_mapped_channel method.
        self.expectThat(
            cm.get_mapped_channel(
                channel_name="latest/stable", architecture="amd64", progressive=False
            ),
            Equals(cm.channel_map[0]),
        )
        self.expectThat(
            cm.get_mapped_channel(
                channel_name="latest/stable", architecture="amd64", progressive=True
            ),
            Equals(cm.channel_map[1]),
        )
        self.assertRaises(
            ValueError,
            cm.get_mapped_channel,
            channel_name="latest/stable",
            architecture="arm64",
            progressive=True,
        )
        self.assertRaises(
            ValueError,
            cm.get_mapped_channel,
            channel_name="latest/stable",
            architecture="i386",
            progressive=True,
        )

        # Test the get_channel_info method.
        self.expectThat(
            cm.get_channel_info("latest/stable"), Equals(cm.snap.channels[0])
        )
        self.assertRaises(ValueError, cm.get_channel_info, "other-track/stable")

        # Test the get_revision method.
        self.expectThat(cm.get_revision(4), Equals(cm.revisions[2]))
        self.assertRaises(ValueError, cm.get_revision, 5)

        # Test the get_existing_architectures method.
        self.expectThat(
            cm.get_existing_architectures(), Equals(set(["arm64", "amd64", "i386"]))
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017,2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from textwrap import dedent

from testtools.matchers import Contains, FileContains

from snapcraft.plugins.v1 import _python

from ._basesuite import PythonBaseTestCase


def _create_site_py(base_dir):
    site_py = os.path.join(base_dir, "usr", "lib", "pythontest", "site.py")
    os.makedirs(os.path.dirname(site_py))
    open(site_py, "w").close()


def _create_user_site_packages(base_dir):
    user_site_dir = os.path.join(base_dir, "lib", "pythontest", "site-packages")
    os.makedirs(user_site_dir)


class SiteCustomizeTestCase(PythonBaseTestCase):
    def setUp(self):
        super().setUp()

        self.expected_sitecustomize = dedent(
            """\
            import site
            import os

            snap_dir = os.getenv("SNAP")
            snapcraft_stage_dir = os.getenv("SNAPCRAFT_STAGE")
            snapcraft_part_install = os.getenv("SNAPCRAFT_PART_INSTALL")

            # Do not include snap_dir during builds as this will include
            # snapcraft's in-snap site directory.
            if snapcraft_stage_dir is not None and snapcraft_part_install is not None:
                site_directories = [snapcraft_stage_dir, snapcraft_part_install]
            else:
                site_directories = [snap_dir]

            for d in site_directories:
                if d:
                    site_dir = os.path.join(d, "lib/pythontest/site-packages")
                    site.addsitedir(site_dir)

            if snap_dir:
                site.ENABLE_USER_SITE = False"""
        )

    def test_generate_sitecustomize_staged(self):
        stage_dir = "stage_dir"
        install_dir = "install_dir"

        # Create the python binary in the staging area
        self._create_python_binary(stage_dir)

        # Create a site.py in both staging and install areas
        _create_site_py(stage_dir)
        _create_site_py(install_dir)

        # Create a user site dir in install area
        _create_user_site_packages(install_dir)

        _python.generate_sitecustomize(
            "test", stage_dir=stage_dir, install_dir=install_dir
        )

        site_path = os.path.join(
            install_dir, "usr", "lib", "pythontest", "sitecustomize.py"
        )
        self.assertThat(site_path, FileContains(self.expected_sitecustomize))

    def test_generate_sitecustomize_installed(self):
        stage_dir = "stage_dir"
        install_dir = "install_dir"

        # Create the python binary in the installed area
        self._create_python_binary(install_dir)

        # Create a site.py in both staging and install areas
        _create_site_py(stage_dir)
        _create_site_py(install_dir)

        # Create a user site dir in install area
        _create_user_site_packages(install_dir)

        _python.generate_sitecustomize(
            "test", stage_dir=stage_dir, install_dir=install_dir
        )

        site_path = os.path.join(
            install_dir, "usr", "lib", "pythontest", "sitecustomize.py"
        )
        self.assertThat(site_path, FileContains(self.expected_sitecustomize))

    def test_generate_sitecustomize_missing_user_site_raises(self):
        stage_dir = "stage_dir"
        install_dir = "install_dir"

        # Create the python binary in the installed area
        self._create_python_binary(install_dir)

        # Create a site.py in both staging and install areas
        _create_site_py(stage_dir)
        _create_site_py(install_dir)

        # Do NOT create a user site dir, and attempt to generate sitecustomize.
        raised = self.assertRaises(
            _python.errors.MissingUserSitePackagesError,
            _python.generate_sitecustomize,
            "test",
            stage_dir=stage_dir,
            install_dir=install_dir,
        )
        self.assertThat(str(raised), Contains("Unable to find user site packages"))

    def test_generate_sitecustomize_missing_site_py_raises(self):
        stage_dir = "stage_dir"
        install_dir = "install_dir"

        # Create the python binary in the staging area
        self._create_python_binary(stage_dir)

        # Create a site.py, but only in install area (not staging area)
        _create_site_py(install_dir)

        # Create a user site dir in install area
        _create_user_site_packages(install_dir)

        raised = self.assertRaises(
            _python.errors.MissingSitePyError,
            _python.generate_sitecustomize,
            "test",
            stage_dir=stage_dir,
            install_dir=install_dir,
        )
        self.assertThat(str(raised), Contains("Unable to find site.py"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2016-2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging
import os
from typing import Any, Dict, Iterable, List, Optional
from urllib.parse import urljoin, urlencode

import requests
from simplejson.scanner import JSONDecodeError

from . import _metadata, constants, errors, http_clients
from ._requests import Requests
from ._status_tracker import StatusTracker
from .v2 import channel_map, releases, validation_sets, whoami


logger = logging.getLogger(__name__)


class DashboardAPI(Requests):
    """The Dashboard API is used to publish and manage snaps.

    This is an interface to query that API which is documented
    at https://dashboard.snapcraft.io/docs/.
    """

    def __init__(self, auth_client: http_clients.AuthClient) -> None:
        self._auth_client = auth_client
        self._root_url = os.environ.get(
            "STORE_DASHBOARD_URL", constants.STORE_DASHBOARD_URL
        )

    def _request(self, method: str, urlpath: str, **kwargs) -> requests.Response:
        url = urljoin(self._root_url, urlpath)
        return self._auth_client.request(method, url, **kwargs)

    def get_macaroon(
        self,
        *,
        acls: Iterable[str],
        packages: Optional[Iterable[Dict[str, str]]] = None,
        channels: Optional[Iterable[str]] = None,
        expires: Optional[Iterable[str]] = None,
    ):
        data: Dict[str, Any] = {"permissions": acls}
        if packages is not None:
            data["packages"] = packages
        if channels is not None:
            data["channels"] = channels
        if expires is not None:
            data["expires"] = expires

        headers = {"Content-Type": "application/json", "Accept": "application/json"}

        if isinstance(self._auth_client, http_clients.CandidClient):
            urlpath = "/api/v2/tokens"
        else:
            urlpath = "/dev/api/acl/"

        response = self.post(urlpath, json=data, headers=headers, auth_header=False)

        if response.ok:
            return response.json()["macaroon"]
        else:
            raise errors.GeneralStoreError("Failed to get macaroon", response)

    def verify_acl(self):
        if not isinstance(self._auth_client, http_clients.UbuntuOneAuthClient):
            raise NotImplementedError("Only supports UbuntuOneAuthClient.")

        response = self.post(
            "/dev/api/acl/verify/",
            json={"auth_data": {"authorization": self._auth_client.auth}},
            headers={"Accept": "application/json"},
            auth_header=False,
        )
        if response.ok:
            return response.json()
        else:
            raise errors.StoreAccountInformationError(response)

    def get_account_information(self) -> Dict[str, Any]:
        response = self.get("/dev/api/account", headers={"Accept": "application/json"})
        if response.ok:
            return response.json()
        else:
            raise errors.StoreAccountInformationError(response)

    def register_key(self, account_key_request):
        data = {"account_key_request": account_key_request}
        response = self.post(
            "/dev/api/account/account-key",
            data=json.dumps(data),
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )
        if not response.ok:
            raise errors.StoreKeyRegistrationError(response)

    def register(
        self, snap_name: str, *, is_private: bool, series: str, store_id: Optional[str]
    ) -> None:
        data = dict(snap_name=snap_name, is_private=is_private, series=series)
        if store_id is not None:
            data["store"] = store_id
        response = self.post(
            "/dev/api/register-name/",
            data=json.dumps(data),
            headers={"Content-Type": "application/json"},
        )
        if not response.ok:
            raise errors.StoreRegistrationError(snap_name, response)

    def snap_upload_precheck(self, snap_name):
        data = {"name": snap_name, "dry_run": True}
        response = self.post(
            "/dev/api/snap-push/",
            data=json.dumps(data),
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )
        if not response.ok:
            raise errors.StoreUploadError(snap_name, response)

    def snap_upload_metadata(
        self,
        snap_name,
        updown_data,
        delta_format=None,
        delta_hash=None,
        source_hash=None,
        target_hash=None,
        built_at=None,
        channels: Optional[List[str]] = None,
    ) -> StatusTracker:
        data = {
            "name": snap_name,
            "series": constants.DEFAULT_SERIES,
            "updown_id": updown_data["upload_id"],
            "binary_filesize": updown_data["binary_filesize"],
            "source_uploaded": updown_data["source_uploaded"],
        }

        if delta_format:
            data["delta_format"] = delta_format
            data["delta_hash"] = delta_hash
            data["source_hash"] = source_hash
            data["target_hash"] = target_hash
        if built_at is not None:
            data["built_at"] = built_at
        if channels is not None:
            data["channels"] = channels
        response = self.post(
            "/dev/api/snap-push/",
            data=json.dumps(data),
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )
        if not response.ok:
            raise errors.StoreUploadError(data["name"], response)

        return StatusTracker(response.json()["status_details_url"])

    def upload_metadata(self, snap_id, snap_name, metadata, force):
        """Upload the metadata to SCA."""
        metadata_handler = _metadata.StoreMetadataHandler(
            request_method=self._request, snap_id=snap_id, snap_name=snap_name,
        )
        metadata_handler.upload(metadata, force)

    def upload_binary_metadata(self, snap_id, snap_name, metadata, force):
        """Upload the binary metadata to SCA."""
        metadata_handler = _metadata.StoreMetadataHandler(
            request_method=self._request, snap_id=snap_id, snap_name=snap_name,
        )
        metadata_handler.upload_binary(metadata, force)

    def snap_release(
        self,
        snap_name,
        revision,
        channels,
        delta_format=None,
        progressive_percentage: Optional[int] = None,
    ):
        data = {"name": snap_name, "revision": str(revision), "channels": channels}
        if delta_format:
            data["delta_format"] = delta_format
        if progressive_percentage is not None:
            data["progressive"] = {
                "percentage": progressive_percentage,
                "paused": False,
            }
        response = self.post(
            "/dev/api/snap-release/",
            data=json.dumps(data),
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )
        if not response.ok:
            raise errors.StoreReleaseError(data["name"], response)

        response_json = response.json()

        return response_json

    def push_assertion(self, snap_id, assertion, endpoint, force):
        if endpoint == "validations":
            data = {"assertion": assertion.decode("utf-8")}
        elif endpoint == "developers":
            data = {"snap_developer": assertion.decode("utf-8")}
        else:
            raise RuntimeError("No valid endpoint")

        url = "/dev/api/snaps/{}/{}".format(snap_id, endpoint)

        # For `snap-developer`, revoking developers will require their uploads
        # to be invalidated.
        if force:
            url = url + "?ignore_revoked_uploads"

        response = self.put(
            url,
            json=data,
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )

        if not response.ok:
            raise errors.StoreValidationError(snap_id, response)
        try:
            response_json = response.json()
        except JSONDecodeError:
            message = (
                "Invalid response from the server when pushing validations: {} {}"
            ).format(response.status_code, response)
            logger.debug(message)
            raise errors.StoreValidationError(
                snap_id, response, message="Invalid response from the server"
            )

        return response_json

    def get_assertion(self, snap_id, endpoint, params=None):
        response = self.get(
            f"/dev/api/snaps/{snap_id}/{endpoint}",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            params=params,
        )
        if not response.ok:
            raise errors.StoreValidationError(snap_id, response)
        try:
            response_json = response.json()
        except JSONDecodeError:
            message = "Invalid response from the server when getting {}: {} {}".format(
                endpoint, response.status_code, response
            )
            logger.debug(message)
            raise errors.StoreValidationError(
                snap_id, response, message="Invalid response from the server"
            )

        return response_json

    def push_snap_build(self, snap_id, snap_build):
        url = f"/dev/api/snaps/{snap_id}/builds"
        data = json.dumps({"assertion": snap_build})
        headers = {
            "Content-Type": "application/json",
        }
        response = self.post(url, data=data, headers=headers)
        if not response.ok:
            raise errors.StoreSnapBuildError(response)

    def snap_status(self, snap_id, series, arch):
        qs = {}
        if series:
            qs["series"] = series
        if arch:
            qs["architecture"] = arch
        url = "/dev/api/snaps/" + snap_id + "/state"
        if qs:
            url += "?" + urlencode(qs)
        response = self.get(
            url,
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )
        if not response.ok:
            raise errors.StoreSnapStatusError(response, snap_id, series, arch)

        response_json = response.json()

        return response_json

    def close_channels(self, snap_id, channel_names):
        url = "/dev/api/snaps/{}/close".format(snap_id)
        data = {"channels": channel_names}
        headers = {"Content-Type": "application/json", "Accept": "application/json"}

        response = self.post(url, data=json.dumps(data), headers=headers)
        if not response.ok:
            raise errors.StoreChannelClosingError(response)

        try:
            results = response.json()
            return results["closed_channels"], results["channel_map_tree"]
        except (JSONDecodeError, KeyError):
            logger.debug(
                "Invalid response from the server on channel closing:\n"
                "{} {}\n{}".format(
                    response.status_code, response.reason, response.content
                )
            )
            raise errors.StoreChannelClosingError(response)

    def sign_developer_agreement(self, latest_tos_accepted=False):
        data = {"latest_tos_accepted": latest_tos_accepted}
        response = self.post(
            "/dev/api/agreement/",
            json=data,
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )

        if not response.ok:
            raise errors.DeveloperAgreementSignError(response)
        return response.json()

    def get_snap_channel_map(self, *, snap_name: str) -> channel_map.ChannelMap:
        response = self.get(
            f"/api/v2/snaps/{snap_name}/channel-map",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )

        if not response.ok:
            raise errors.StoreSnapChannelMapError(snap_name=snap_name)

        return channel_map.ChannelMap.unmarshal(response.json())

    def get_snap_releases(self, *, snap_name: str) -> releases.Releases:
        response = self.get(
            f"/api/v2/snaps/{snap_name}/releases",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )

        if not response.ok:
            raise errors.StoreSnapChannelMapError(snap_name=snap_name)

        return releases.Releases.unmarshal(response.json())

    def whoami(self) -> whoami.WhoAmI:
        response = self.get(
            "/api/v2/tokens/whoami",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
        )

        if not response.ok:
            raise errors.GeneralStoreError(message="whoami failed.", response=response)

        return whoami.WhoAmI.unmarshal(response.json())

    def post_validation_sets_build_assertion(
        self, validation_sets_data: Dict[str, Any]
    ) -> validation_sets.BuildAssertion:
        url = "/api/v2/validation-sets/build-assertion"
        response = self.post(
            url,
            headers={"Accept": "application/json", "Content-Type": "application/json"},
            json=validation_sets_data,
        )

        if not response.ok:
            raise errors.StoreValidationSetsError(response)

        return validation_sets.BuildAssertion.unmarshal(response.json())

    def post_validation_sets(
        self, signed_validation_sets: bytes
    ) -> validation_sets.ValidationSets:
        url = "/api/v2/validation-sets"
        response = self.post(
            url,
            headers={
                "Accept": "application/json",
                "Content-Type": "application/x.ubuntu.assertion",
            },
            data=signed_validation_sets,
        )

        if not response.ok:
            raise errors.StoreValidationSetsError(response)

        return validation_sets.ValidationSets.unmarshal(response.json())

    def get_validation_sets(
        self, *, name: Optional[str], sequence: Optional[str]
    ) -> validation_sets.ValidationSets:
        url = "/api/v2/validation-sets"
        if name is not None:
            url += "/" + name
        params = dict()
        if sequence is not None:
            params["sequence"] = sequence

        response = self.get(url, headers={"Accept": "application/json"}, params=params)

        if not response.ok:
            raise errors.StoreValidationSetsError(response)

        return validation_sets.ValidationSets.unmarshal(response.json())



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import itertools
import logging
import os
import pathlib
import subprocess
import sys
import time
import typing
from typing import List, Optional, Union

import click
import progressbar

from snapcraft import file_utils
from snapcraft.internal import (
    build_providers,
    deprecations,
    errors,
    indicators,
    lifecycle,
    project_loader,
    steps,
)
from snapcraft.internal.repo import ua_manager
from snapcraft.project._sanity_checks import conduct_project_sanity_check

from . import echo
from ._command import SnapcraftProjectCommand
from ._errors import TRACEBACK_HOST, TRACEBACK_MANAGED
from ._options import (
    add_provider_options,
    apply_host_provider_flags,
    get_build_provider,
    get_build_provider_flags,
    get_project,
)

logger = logging.getLogger(__name__)


if typing.TYPE_CHECKING:
    from snapcraft.internal.project import Project  # noqa: F401


# TODO: when snap is a real step we can simplify the arguments here.
def _execute(  # noqa: C901
    step: steps.Step,
    parts: str,
    pack_project: bool = False,
    output: Optional[str] = None,
    shell: bool = False,
    shell_after: bool = False,
    setup_prime_try: bool = False,
    ua_token: Optional[str] = None,
    **kwargs,
) -> "Project":
    # Cleanup any previous errors.
    _clean_provider_error()

    build_provider = get_build_provider(**kwargs)
    build_provider_flags = get_build_provider_flags(
        build_provider, ua_token=ua_token, **kwargs
    )
    apply_host_provider_flags(build_provider_flags)

    is_managed_host = build_provider == "managed-host"

    # Temporary fix to ignore target_arch.
    if kwargs.get("target_arch") is not None and build_provider in ["multipass", "lxd"]:
        echo.warning(
            "Ignoring '--target-arch' flag.  This flag requires --destructive-mode and is unsupported with Multipass and LXD build providers."
        )
        kwargs.pop("target_arch")

    project = get_project(is_managed_host=is_managed_host, **kwargs)
    conduct_project_sanity_check(project, **kwargs)

    project_path = pathlib.Path(project._project_dir)
    if project_path.name in ["build-aux", "snap"]:
        echo.warning(
            f"Snapcraft is running in directory {project_path.name!r}.  If this is the snap assets directory, please run snapcraft from {project_path.parent}."
        )

    if build_provider in ["host", "managed-host"]:
        with ua_manager.ua_manager(ua_token):
            project_config = project_loader.load_config(project)
            lifecycle.execute(step, project_config, parts)
            if pack_project:
                _pack(
                    project.prime_dir,
                    compression=project._snap_meta.compression,
                    output=output,
                )
    else:
        build_provider_class = build_providers.get_provider_for(build_provider)
        try:
            build_provider_class.ensure_provider()
        except build_providers.errors.ProviderNotFound as provider_error:
            if provider_error.prompt_installable:
                if echo.is_tty_connected() and echo.confirm(
                    "Support for {!r} needs to be set up. "
                    "Would you like to do it now?".format(provider_error.provider)
                ):
                    build_provider_class.setup_provider(echoer=echo)
                else:
                    raise provider_error
            else:
                raise provider_error

        with build_provider_class(
            project=project, echoer=echo, build_provider_flags=build_provider_flags
        ) as instance:
            try:
                if shell:
                    # shell means we want to do everything right up to the previous
                    # step and then go into a shell instead of the requested step.
                    # the "snap" target is a special snowflake that has not made its
                    # way to be a proper step.
                    previous_step = None
                    if pack_project:
                        previous_step = steps.PRIME
                    elif step > steps.PULL:
                        previous_step = step.previous_step()
                    # steps.PULL is the first step, so we would directly shell into it.
                    if previous_step:
                        instance.execute_step(previous_step)
                elif pack_project:
                    instance.pack_project(output=output)
                elif setup_prime_try:
                    instance.expose_prime()
                    instance.execute_step(step)
                else:
                    instance.execute_step(step)
            except Exception:
                _retrieve_provider_error(instance)
                if project.debug:
                    instance.shell()
                else:
                    echo.warning(
                        "Run the same command again with --debug to shell into the environment "
                        "if you wish to introspect this failure."
                    )
                    raise
            else:
                if shell or shell_after:
                    instance.shell()
    return project


def _run_pack(snap_command: List[Union[str, pathlib.Path]]) -> str:
    ret = None
    stdout = ""
    stderr = ""
    with subprocess.Popen(
        snap_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    ) as proc:
        if indicators.is_dumb_terminal():
            echo.info("Snapping...")
            ret = proc.wait()
        else:
            message = "\033[0;32mSnapping \033[0m"
            progress_indicator = progressbar.ProgressBar(
                widgets=[message, progressbar.AnimatedMarker()],
                # From progressbar.ProgressBar.update(...).
                maxval=progressbar.UnknownLength,
            )
            progress_indicator.start()
            for counter in itertools.count():
                progress_indicator.update(counter)
                time.sleep(0.2)
                ret = proc.poll()
                if ret is not None:
                    break
            progress_indicator.finish()

        if proc.stdout is not None:
            stdout = proc.stdout.read().decode()
        if proc.stderr is not None:
            stderr = proc.stderr.read().decode()
        logger.debug(f"stdout: {stdout} | stderr: {stderr}")

    if ret != 0:
        raise RuntimeError(
            f"Failed to create snap, snap command failed:\nstdout:\n{stdout}\nstderr:\n{stderr}"
        )

    try:
        snap_filename = stdout.split(":")[1].strip()
    except IndexError:
        logger.debug("Failed to parse snap pack outpout: {stdout}")
        snap_filename = stdout

    return snap_filename


def _pack(
    directory: str, *, compression: Optional[str] = None, output: Optional[str]
) -> None:
    """Pack a snap.

    :param directory: directory to snap
    :param compression: compression type to use, None for defaults
    :param output: Output may either be:
        (1) a directory path to output snaps to
        (2) an explicit file path to output snap to
        (3) unpsecified/None to output to current (project) directory
    """
    output_file = None
    output_dir = None

    if output:
        output_path = pathlib.Path(output)
        output_parent = output_path.parent
        if output_path.is_dir():
            output_dir = str(output_path)
        elif output_parent and output_parent != pathlib.Path("."):
            output_dir = str(output_parent)
            output_file = output_path.name
        else:
            output_file = output

    snap_path = file_utils.get_host_tool_path(command_name="snap", package_name="snapd")

    command: List[Union[str, pathlib.Path]] = [snap_path, "pack"]
    # When None, just use snap pack's default settings.
    if compression is not None:
        command.extend(["--compression", compression])

    if output_file is not None:
        command.extend(["--filename", output_file])

    command.append(directory)

    if output_dir is not None:
        command.append(output_dir)

    logger.debug(f"Running pack command: {command}")
    snap_filename = _run_pack(command)
    echo.info(f"Snapped {snap_filename}")


def _clean_provider_error() -> None:
    if os.path.isfile(TRACEBACK_HOST):
        try:
            os.remove(TRACEBACK_HOST)
        except Exception as e:
            logger.debug("can't remove error file: {}", str(e))


def _retrieve_provider_error(instance) -> None:
    try:
        instance.pull_file(TRACEBACK_MANAGED, TRACEBACK_HOST, delete=True)
    except Exception as e:
        logger.debug("can't retrieve error file: {}", str(e))


@click.group()
@add_provider_options()
@click.pass_context
def lifecyclecli(ctx, **kwargs):
    pass


@lifecyclecli.command()
def init():
    """Initialize a snapcraft project."""
    snapcraft_yaml_path = lifecycle.init()
    echo.info("Created {}.".format(snapcraft_yaml_path))
    echo.wrapped(
        "Go to https://docs.snapcraft.io/the-snapcraft-format/8337 for more "
        "information about the snapcraft.yaml format."
    )


@lifecyclecli.command(cls=SnapcraftProjectCommand)
@click.pass_context
@add_provider_options()
@click.argument("parts", nargs=-1, metavar="<part>...", required=False)
def pull(ctx, parts, **kwargs):
    """Download or retrieve artifacts defined for a part.

    \b
    Examples:
        snapcraft pull
        snapcraft pull my-part1 my-part2

    """
    _execute(steps.PULL, parts, **kwargs)


@lifecyclecli.command(cls=SnapcraftProjectCommand)
@add_provider_options()
@click.argument("parts", nargs=-1, metavar="<part>...", required=False)
def build(parts, **kwargs):
    """Build artifacts defined for a part.

    \b
    Examples:
        snapcraft build
        snapcraft build my-part1 my-part2

    """
    _execute(steps.BUILD, parts, **kwargs)


@lifecyclecli.command(cls=SnapcraftProjectCommand)
@add_provider_options()
@click.argument("parts", nargs=-1, metavar="<part>...", required=False)
def stage(parts, **kwargs):
    """Stage the part's built artifacts into the common staging area.

    \b
    Examples:
        snapcraft stage
        snapcraft stage my-part1 my-part2

    """
    _execute(steps.STAGE, parts, **kwargs)


@lifecyclecli.command(cls=SnapcraftProjectCommand)
@add_provider_options()
@click.argument("parts", nargs=-1, metavar="<part>...", required=False)
def prime(parts, **kwargs):
    """Final copy and preparation for the snap.

    \b
    Examples:
        snapcraft prime
        snapcraft prime my-part1 my-part2

    """
    _execute(steps.PRIME, parts, **kwargs)


@lifecyclecli.command("try")
@add_provider_options()
def try_command(**kwargs):
    """Try a snap on the host, priming if necessary.

    This feature only works on snap enabled systems.

    \b
    Examples:
        snapcraft try

    """
    project = _execute(steps.PRIME, [], setup_prime_try=True, **kwargs)
    # project.prime_dir here points to the on-host prime directory.
    echo.info("You can now run `snap try {}`.".format(project.prime_dir))


@lifecyclecli.command(cls=SnapcraftProjectCommand)
@add_provider_options()
@click.argument("directory", required=False)
@click.option("--output", "-o", help="path to the resulting snap.")
def snap(directory, output, **kwargs):
    """Create a snap.

    \b
    Examples:
        snapcraft snap
        snapcraft snap --output renamed-snap.snap

    If you want to snap a directory, you should use the pack command
    instead.
    """
    if directory:
        deprecations.handle_deprecation_notice("dn6")
        _pack(directory, output=output)
    else:
        _execute(steps.PRIME, parts=tuple(), pack_project=True, output=output, **kwargs)


@lifecyclecli.command(cls=SnapcraftProjectCommand)
@click.argument("directory")
@click.option("--output", "-o", help="path to the resulting snap.")
def pack(directory, output, **kwargs):
    """Create a snap from a directory holding a valid snap.

    The layout of <directory> should contain a valid meta/snap.yaml in
    order to be a valid snap.

    \b
    Examples:
        snapcraft pack my-snap-directory
        snapcraft pack my-snap-directory --output renamed-snap.snap

    """
    _pack(directory, output=output)


@lifecyclecli.command(cls=SnapcraftProjectCommand)
@click.pass_context
@add_provider_options()
@click.argument("parts", nargs=-1, metavar="<part>...", required=False)
@click.option("--unprime", is_flag=True, required=False, hidden=True)
@click.option("--step", "-s", required=False, hidden=True)
def clean(ctx, parts, unprime, step, **kwargs):
    """Remove a part's assets.

    \b
    Examples:
        snapcraft clean
        snapcraft clean my-part
    """
    # This option is only valid in legacy.
    if step:
        option = "--step" if "--step" in ctx.obj["argv"] else "-s"
        raise click.BadOptionUsage(option, "no such option: {}".format(option))

    build_provider = get_build_provider(**kwargs)
    build_provider_flags = get_build_provider_flags(build_provider, **kwargs)
    apply_host_provider_flags(build_provider_flags)

    is_managed_host = build_provider == "managed-host"

    # Temporary fix to ignore target_arch, silently for clean.
    if "target_arch" in kwargs and build_provider in ["multipass", "lxd"]:
        kwargs.pop("target_arch")

    try:
        project = get_project(is_managed_host=is_managed_host)
    except errors.ProjectNotFoundError:
        # Fresh environment, nothing to clean.
        return

    if unprime and not is_managed_host:
        raise click.BadOptionUsage("--unprime", "no such option: --unprime")

    if build_provider in ["host", "managed-host"]:
        step = steps.PRIME if unprime else None
        lifecycle.clean(project, parts, step)
    else:
        build_provider_class = build_providers.get_provider_for(build_provider)
        if parts:
            with build_provider_class(
                project=project, echoer=echo, build_provider_flags=build_provider_flags
            ) as instance:
                instance.clean_parts(part_names=parts)
        else:
            build_provider_class(project=project, echoer=echo).clean_project()
            # Clear the prime directory on the host, unless on Windows.
            if sys.platform != "win32":
                lifecycle.clean(project, parts, steps.PRIME)


if __name__ == "__main__":
    lifecyclecli.main()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import platform
import sys

_32BIT_USERSPACE_ARCHITECTURE = {
    "aarch64": "armv7l",
    "armv8l": "armv7l",
    "ppc64le": "ppc",
    "x86_64": "i686",
}


_ARCH_TRANSLATIONS = {
    "armv7l": {"deb": "armhf", "triplet": "arm-linux-gnueabihf"},
    "aarch64": {"deb": "arm64", "triplet": "aarch64-linux-gnu"},
    "i686": {"deb": "i386", "triplet": "i386-linux-gnu"},
    "ppc64le": {"deb": "ppc64el", "triplet": "powerpc64le-linux-gnu"},
    "ppc": {"deb": "powerpc", "triplet": "powerpc-linux-gnu"},
    "x86_64": {"deb": "amd64", "triplet": "x86_64-linux-gnu"},
    "s390x": {"deb": "s390x", "triplet": "s390x-linux-gnu"},
}


_WINDOWS_TRANSLATIONS = {"AMD64": "x86_64"}


def get_deb_arch():
    return _ARCH_TRANSLATIONS[_get_platform_architecture()]["deb"]


def get_arch_triplet():
    return _ARCH_TRANSLATIONS[_get_platform_architecture()]["triplet"]


def _get_platform_architecture():
    architecture = platform.machine()

    # Translate the windows architectures we know of to architectures
    # we can work with.
    if sys.platform == "win32":
        architecture = _WINDOWS_TRANSLATIONS.get(architecture)

    if platform.architecture()[0] == "32bit":
        userspace = _32BIT_USERSPACE_ARCHITECTURE.get(architecture)
        if userspace:
            architecture = userspace

    return architecture



import base64
import json
import os
import pathlib
from typing import Optional, TextIO
from urllib.parse import urlparse

import requests
import macaroonbakery._utils as utils
from macaroonbakery import bakery, httpbakery
from xdg import BaseDirectory

from snapcraft.storeapi import constants
from . import agent, _config, _http_client


class CandidConfig(_config.Config):
    """Hold configuration options in sections.

    There can be two sections for the sso related credentials: production and
    staging. This is governed by the STORE_DASHBOARD_URL environment
    variable. Other sections are ignored but preserved.

    """

    def _get_section_name(self) -> str:
        url = os.getenv("STORE_DASHBOARD_URL", constants.STORE_DASHBOARD_URL)
        return urlparse(url).netloc

    def _get_config_path(self) -> pathlib.Path:
        return pathlib.Path(BaseDirectory.save_config_path("snapcraft")) / "candid.cfg"


class CandidClient(_http_client.Client):
    @classmethod
    def has_credentials(cls) -> bool:
        return not CandidConfig().is_section_empty()

    @property
    def _macaroon(self) -> Optional[str]:
        return self._conf.get("macaroon")

    @_macaroon.setter
    def _macaroon(self, macaroon: str) -> None:
        self._conf.set("macaroon", macaroon)
        if self._conf_save:
            self._conf.save()

    @property
    def _auth(self) -> Optional[str]:
        return self._conf.get("auth")

    @_auth.setter
    def _auth(self, auth: str) -> None:
        self._conf.set("auth", auth)
        if self._conf_save:
            self._conf.save()

    def __init__(
        self,
        *,
        user_agent: str = agent.get_user_agent(),
        bakery_client=httpbakery.Client(),
    ) -> None:
        super().__init__(user_agent=user_agent)

        self.bakery_client = bakery_client
        self._conf = CandidConfig()
        self._conf_save = True

    def _login(self, macaroon: str) -> None:
        bakery_macaroon = bakery.Macaroon.from_dict(json.loads(macaroon))
        discharges = bakery.discharge_all(
            bakery_macaroon, self.bakery_client.acquire_discharge
        )

        # serialize macaroons the bakery-way
        discharged_macaroons = (
            "[" + ",".join(map(utils.macaroon_to_json_string, discharges)) + "]"
        )

        self._auth = base64.urlsafe_b64encode(
            utils.to_bytes(discharged_macaroons)
        ).decode("ascii")
        self._macaroon = macaroon

    def login(
        self,
        *,
        macaroon: Optional[str] = None,
        config_fd: Optional[TextIO] = None,
        save: bool = True,
    ) -> None:
        self._conf_save = save
        if macaroon is not None:
            self._login(macaroon)
        elif config_fd is not None:
            self._conf.load(config_fd=config_fd)
            if save:
                self._conf.save()
        else:
            raise RuntimeError("Logic Error")

    def request(
        self, method, url, params=None, headers=None, auth_header=True, **kwargs
    ) -> requests.Response:
        if headers and auth_header:
            headers["Macaroons"] = self._auth
        elif auth_header:
            headers = {"Macaroons": self._auth}

        response = super().request(
            method, url, params=params, headers=headers, **kwargs
        )

        if not response.ok and response.status_code == 401:
            self.login(macaroon=self._macaroon)

            response = super().request(
                method, url, params=params, headers=headers, **kwargs
            )

        return response

    def export_login(self, *, config_fd: TextIO, encode: bool):
        self._conf.save(config_fd=config_fd, encode=encode)

    def logout(self) -> None:
        self._conf.clear()
        self._conf.save()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snapcraft.internal.meta import errors


class TestErrorFormatting:

    scenarios = (
        (
            "MissingSnapcraftYamlKeysError",
            {
                "exception_class": errors.MissingSnapcraftYamlKeysError,
                "kwargs": {"keys": ["test-key1", "test-key2"]},
                "expected_message": (
                    "Failed to generate snap metadata: "
                    "Missing required key(s) in snapcraft.yaml: "
                    "'test-key1' and 'test-key2'. Either specify the missing "
                    "key(s), or use 'adopt-info' to get them from a part."
                ),
            },
        ),
        (
            "AdoptedPartMissingError",
            {
                "exception_class": errors.AdoptedPartMissingError,
                "kwargs": {"part": "test-part"},
                "expected_message": (
                    "Failed to generate snap metadata: "
                    "'adopt-info' refers to a part named 'test-part', but it is "
                    "not defined in the 'snapcraft.yaml' file."
                ),
            },
        ),
        (
            "AdoptedPartNotParsingInfo",
            {
                "exception_class": errors.AdoptedPartNotParsingInfo,
                "kwargs": {"part": "test-part"},
                "expected_message": (
                    "Failed to generate snap metadata: "
                    "'adopt-info' refers to part 'test-part', but that part is "
                    "lacking the 'parse-info' property."
                ),
            },
        ),
        (
            "AmbiguousPassthroughKeyError",
            {
                "exception_class": errors.AmbiguousPassthroughKeyError,
                "kwargs": {"keys": ["key1", "key2"]},
                "expected_message": (
                    "Failed to generate snap metadata: "
                    "The following keys are specified in their regular location "
                    "as well as in passthrough: 'key1' and 'key2'. "
                    "Remove duplicate keys."
                ),
            },
        ),
        (
            "InvalidAppCommandError",
            {
                "exception_class": errors.InvalidAppCommandError,
                "kwargs": {"command": "test-command", "app_name": "test-app"},
                "expected_message": (
                    "Failed to generate snap metadata: "
                    "The specified command 'test-command' defined in the app "
                    "'test-app' does not exist or is not executable.\n"
                    "Ensure that 'test-command' is installed with the correct path."
                ),
            },
        ),
        (
            "InvalidAppCommandNotFound",
            {
                "exception_class": errors.InvalidAppCommandNotFound,
                "kwargs": {"command": "test-command", "app_name": "test-app"},
                "expected_message": (
                    "Failed to generate snap metadata: "
                    "The specified command 'test-command' defined in the app "
                    "'test-app' does not exist.\n"
                    "Ensure that 'test-command' is installed with the correct path."
                ),
            },
        ),
        (
            "InvalidAppCommandNotExecutable",
            {
                "exception_class": errors.InvalidAppCommandNotExecutable,
                "kwargs": {"command": "test-command", "app_name": "test-app"},
                "expected_message": (
                    "Failed to generate snap metadata: "
                    "The specified command 'test-command' defined in the app "
                    "'test-app' is not executable."
                ),
            },
        ),
        (
            "InvalidAppCommandFormatError",
            {
                "exception_class": errors.InvalidAppCommandFormatError,
                "kwargs": {"command": "test-command", "app_name": "test-app"},
                "expected_message": (
                    "Failed to generate snap metadata: "
                    "The specified command 'test-command' defined in the app "
                    "'test-app' does not match the pattern expected by snapd.\n"
                    "The command must consist only of alphanumeric characters, spaces, "
                    "and the following special characters: / . _ # : $ -"
                ),
            },
        ),
        (
            "InvalidCommandChainError",
            {
                "exception_class": errors.InvalidCommandChainError,
                "kwargs": {"item": "test-chain", "app_name": "test-app"},
                "expected_message": (
                    "Failed to generate snap metadata: "
                    "The command-chain item 'test-chain' defined in the app 'test-app' "
                    "does not exist or is not executable.\n"
                    "Ensure that 'test-chain' is relative to the prime directory."
                ),
            },
        ),
        (
            "InvalidDesktopFileError",
            {
                "exception_class": errors.InvalidDesktopFileError,
                "kwargs": {"filename": "test-file", "message": "test-message"},
                "expected_message": (
                    "Failed to generate desktop file: "
                    "Invalid desktop file 'test-file': test-message."
                ),
            },
        ),
    )

    def test_error_formatting(self, exception_class, kwargs, expected_message):
        assert str(exception_class(**kwargs)) == expected_message


class TestSnapcraftException:

    scenarios = (
        (
            "GradeDevelRequiredError",
            {
                "exception_class": errors.GradeDevelRequiredError,
                "kwargs": dict(set_grade="stable"),
                "expected_brief": "Snap 'grade' was set to 'stable' but must be 'devel'.",
                "expected_resolution": "Set 'grade' to 'devel' or use a stable base for this snap.",
                "expected_details": None,
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
        (
            "SystemUsernamesValidationError",
            {
                "exception_class": errors.SystemUsernamesValidationError,
                "kwargs": dict(name="user1", message="invalid ..."),
                "expected_brief": "Invalid system-usernames for 'user1': invalid ...",
                "expected_resolution": "Please configure system user according to documentation.",
                "expected_details": None,
                "expected_docs_url": "https://forum.snapcraft.io/t/system-usernames/13386",
                "expected_reportable": False,
            },
        ),
    )

    def test_snapcraft_exception_handling(
        self,
        exception_class,
        kwargs,
        expected_brief,
        expected_resolution,
        expected_details,
        expected_docs_url,
        expected_reportable,
    ):
        exception = exception_class(**kwargs)
        assert exception.get_brief() == expected_brief
        assert exception.get_resolution() == expected_resolution
        assert exception.get_details() == expected_details
        assert exception.get_docs_url() == expected_docs_url
        assert exception.get_reportable() == expected_reportable


def test_PackageRepositoriesValidationError():
    error = errors.PackageRepositoryValidationError(
        url="http://foo", brief="some error", resolution="some way to fix"
    )

    assert (
        error.get_brief() == "Invalid package-repository for 'http://foo': some error"
    )
    assert error.get_resolution() == "some way to fix"
    assert error.get_details() is None
    assert error.get_docs_url() == "https://snapcraft.io/docs/package-repositories"
    assert error.get_reportable() is False


def test_PackageRepositoriesValidationError_no_resolution():
    error = errors.PackageRepositoryValidationError(
        url="http://foo", brief="some error"
    )

    assert (
        error.get_resolution()
        == "You can verify package repository configuration according to the referenced documentation."
    )
